# Copyright 2022 NNAISENSE SA
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""This module contains logging utilities."""

import os
import pathlib
import pickle
import weakref
from copy import deepcopy
from datetime import datetime
from typing import Any, Iterable, Optional, Union

import torch
from torch import nn

from .algorithms.searchalgorithm import SearchAlgorithm
from .core import Problem, Solution
from .neuroevolution.net import device_of_module
from .tools import ReadOnlyTensor, clone, is_dtype_object, is_sequence

try:
    import pandas
except ImportError:
    pandas = None


try:
    import sacred
except ImportError:
    sacred = None


try:
    import mlflow
except ImportError:
    mlflow = None


try:
    import neptune.new as neptune
except ImportError:
    neptune = None


class Logger:
    """Base class for all logging classes."""

    def __init__(self, searcher: SearchAlgorithm, *, interval: int = 1, after_first_step: bool = False):
        """`__init__(...)`: Initialize the Logger.

        Args:
            searcher: The evolutionary algorithm instance whose progress
                is to be logged.
            interval: Expected as an integer n.
                Logging is to be done at every n iterations.
            after_first_step: Expected as a boolean.
                Meaningful only if interval is set as an integer greater
                than 1. Let us suppose that interval is set as 10.
                If after_first_step is False (which is the default),
                then the logging will be done at steps 10, 20, 30, and so on.
                On the other hand, if after_first_step is True,
                then the logging will be done at steps 1, 11, 21, 31, and so
                on.
        """
        searcher.log_hook.append(self)
        self._interval = int(interval)
        self._after_first_step = bool(after_first_step)
        self._steps_count = 0

    def __call__(self, status: dict):
        if self._after_first_step:
            n = self._steps_count
            self._steps_count += 1
        else:
            self._steps_count += 1
            n = self._steps_count

        if (n % self._interval) == 0:
            self._log(self._filter(status))

    def _filter(self, status: dict) -> dict:
        return status

    def _log(self, status: dict):
        raise NotImplementedError


class PicklingLogger(Logger):
    """
    A logger which periodically pickles the current result of the search.

    The pickled data includes the current center solution and the best solution
    (if available). If the problem being solved is a reinforcement learning
    task, then the pickled data also includes the observation normalization
    data and the policy.
    """

    def __init__(
        self,
        searcher: SearchAlgorithm,
        *,
        interval: int,
        directory: Optional[Union[str, pathlib.Path]] = None,
        prefix: Optional[str] = None,
        zfill: int = 6,
        items_to_save: Union[str, Iterable] = (
            "center",
            "best",
            "pop_best",
            "median_eval",
            "mean_eval",
            "pop_best_eval",
        ),
        make_policy_from: Optional[str] = None,
        after_first_step: bool = False,
        verbose: bool = True,
    ):
        """
        `__init__(...)`: Initialize the PicklingLogger.

        Args:
            searcher: The evolutionary algorithm instance whose progress
                is to be pickled.
            interval: Expected as an integer n.
                Pickling is to be done at every n iterations.
            directory: The directory into which the pickle files will be
                placed. If given as None, the current directory will be
                used.
            prefix: The prefix to be used when naming the pickle file.
                If left as None, then the prefix will contain the date,
                time, name of the class of the problem object, and the PID
                of the current process.
            zfill: When naming the pickle file, the generation number will
                also be used. This `zfill` argument is used to determine
                the length of the number of generations string.
                For example, if the current number of generations is
                150, and zfill is 6 (which is the default), then the
                number of generations part of the file name will look
                like this: 'generation000150', the numeric part of the
                string taking 6 characters.
            items_to_save: Items to save from the status dictionary.
                This can be a string, or a sequence of strings.
            make_policy_from: If given as a string, then the solution
                represented by this key in the status dictionary will be
                taken, and converted a PyTorch module with the help of the
                problem's `to_policy(...)` method, and then will be added
                into the pickle file. For example, if this argument is
                given as "center", then the center solution will be converted
                to a policy and saved.
                If this argument is left as None, then, this logger will
                first try to obtain the center solution. If there is no
                center solution, then the logger will try to obtain the
                'pop_best' solution. If that one does not exist either, then
                an error will be raised, and the user will be requested to
                explicitly state a value for the argument `make_policy_from`.
                If the problem object does not provide a `to_policy(...)`,
                method, then this configuration will be ignored, and no
                policy will be saved.
            after_first_step: Expected as a boolean.
                Meaningful only if interval is set as an integer greater
                than 1. Let us suppose that interval is set as 10.
                If after_first_step is False (which is the default),
                then the logging will be done at steps 10, 20, 30, and so on.
                On the other hand, if after_first_step is True,
                then the logging will be done at steps 1, 11, 21, 31, and so
                on.
            verbose: If set as True, then a message will be printed out to
                the standard output every time the pickling happens.
        """
        # Call the initializer of the superclass.
        super().__init__(searcher, interval=interval, after_first_step=after_first_step)

        # Each Logger register itself to the search algorithm's log hook. Additionally, in the case of PicklingLogger,
        # we add this object's final saving method to the search algorithm's end-of-run hook.
        # This is to make sure that the latest generation's result is saved, even when the last generation number
        # does not coincide with the saving period.
        searcher.end_of_run_hook.append(self._final_save)

        # Store a weak reference to the search algorithm.
        self._searcher_ref = weakref.ref(searcher)

        # Store the item keys as a tuple of strings.
        if isinstance(items_to_save, str):
            self._items_to_save = (items_to_save,)
        else:
            self._items_to_save = tuple(items_to_save)

        # Store the status key that will be used to get the current solution for making the policy.
        self._make_policy_from = None if make_policy_from is None else str(make_policy_from)

        if prefix is None:
            # If a file name prefix is not given by the user, then we prepare one using the current date and time,
            # name of the problem type, and the PID of the current process.
            strnow = datetime.now().strftime("%Y-%m-%d-%H.%M.%S")
            probname = type(searcher.problem).__name__
            strpid = str(os.getpid())
            self._prefix = f"{probname}_{strnow}_{strpid}"
        else:
            # If there is a file name prefix given by the user, then we use that name.
            self._prefix = str(prefix)

        if directory is None:
            # If a directory name is not given by the user, then we store the directory name as None.
            self._directory = None
        else:
            # If a directory name is given by the user, then we store it as a string, and make sure that it exists.
            self._directory = str(directory)
            os.makedirs(self._directory, exist_ok=True)

        self._verbose = bool(verbose)
        self._zfill = int(zfill)

        self._last_generation: Optional[int] = None
        self._last_file_name: Optional[str] = None

    def _within_dir(self, fname: Union[str, pathlib.Path]) -> str:
        fname = str(fname)
        if self._directory is not None:
            fname = os.path.join(self._directory, fname)
        return fname

    @classmethod
    def _as_cpu_tensor(cls, x: Any) -> Any:
        if isinstance(x, Solution):
            x = cls._as_cpu_tensor(x.values)
        elif isinstance(x, torch.Tensor):
            with torch.no_grad():
                x = x.cpu().clone()
            if isinstance(x, ReadOnlyTensor):
                x = x.as_subclass(torch.Tensor)
        return x

    def _log(self, status: dict):
        self.save()

    def _final_save(self, status: dict):
        # Get the stored search algorithm
        searcher: Optional[SearchAlgorithm] = self._searcher_ref()

        if searcher is not None:
            # If the search algorithm object is still alive, then we check its generation number
            current_gen = searcher.step_count

            if (self._last_generation is None) or (current_gen > self._last_generation):
                # If there was not a save at all, or the latest save was for a previous generation,
                # then save the current status of the search.
                self.save()

    def save(self, fname: Optional[Union[str, pathlib.Path]] = None) -> str:
        """
        Pickle the current status of the evolutionary search.

        If this PicklingLogger was initialized with a `directory` argument,
        then the pickle file will be put into that directory.

        Args:
            fname: The name of the pickle file to be created.
                This can be left as None if the user wishes this PicklingLogger
                to determine the name of the file.
                Alternatively, an explicit name for this file can be specified
                via this argument.
        Returns:
            The name of the pickle file that was generated, as a string.
        """
        # Get the stored search algorithm
        searcher: Optional[SearchAlgorithm] = self._searcher_ref()

        if searcher is not None:
            # If the search algorithm object is still alive, then take the problem object from it.
            problem: Problem = searcher.problem

            # Initialize the data dictionary that will contain the objects to be put into the pickle file.
            data = {}

            for item_to_save in self._items_to_save:
                # For each item key, get the object with that key from the algorithm's status dictionary, and then put
                # that object into our data dictionary.
                if item_to_save in searcher.status:
                    data[item_to_save] = self._as_cpu_tensor(searcher.status[item_to_save])

            if (
                hasattr(problem, "observation_normalization")
                and hasattr(problem, "get_observation_stats")
                and problem.observation_normalization
            ):
                # If the problem object has observation normalization, then put the normalizer object to the data
                # dictionary.
                data["obs_stats"] = problem.get_observation_stats().to("cpu")

            if hasattr(problem, "to_policy"):
                # If the problem object has the method `to_policy(...)`, then we will generate a policy from the
                # current solution, and put that policy into the data dictionary.

                if self._make_policy_from is None:
                    # If the user did not specify the status key of the solution that will be converted to a policy,
                    # then, we first check if the search algorithm has a "center" item, and if not, we check if it
                    # has a "pop_best" item.

                    if "center" in searcher.status:
                        # The status key of the search algorithm has "center". We declare "center" as the key of the
                        # current solution.
                        policy_key = "center"
                    elif "pop_best" in searcher.status:
                        # The status key of the search algorithm has "pop_best". We declare "pop_best" as the key of
                        # the current solution.
                        policy_key = "pop_best"
                    else:
                        # The status key of the search algorithm contains neither a "center" solution and nor a
                        # "pop_best" solution. So, we raise an error.
                        raise ValueError(
                            "PicklingLogger did not receive an explicit value for its `make_policy_from` argument."
                            " The status dictionary of the search algorithm has neither 'center' nor 'pop_best'."
                            " Therefore, it is not clear which status item is to be used for making a policy."
                            " Please try instantiating a PicklingLogger with an explicit `make_policy_from` value."
                        )
                else:
                    # This is the case where the user explicitly specified (via a status key) which solution will be
                    # used for making the current policy. We declare that key as the key to use.
                    policy_key = self._make_policy_from

                # Get the solution.
                policy_solution = searcher.status[policy_key]

                # Make a policy from the solution
                policy = problem.to_policy(policy_solution)

                if isinstance(policy, nn.Module) and (device_of_module(policy) != torch.device("cpu")):
                    # If the created policy is a PyTorch module, and this module is not on the cpu, then we make
                    # a copy of this module on the cpu.
                    policy = clone(policy).to("cpu")

                # We put this policy into our data dictionary.
                data["policy"] = policy

                # We put the datetime-related information
                first_step_datetime = searcher.first_step_datetime
                if first_step_datetime is not None:
                    now = datetime.now()
                    elapsed = now - first_step_datetime
                    data["beginning_time"] = first_step_datetime
                    data["now"] = now
                    data["elapsed"] = elapsed

            if fname is None:
                # If a file name was not given, then we generate one.
                num_gens = str(searcher.step_count).zfill(self._zfill)
                fname = f"{self._prefix}_generation{num_gens}.pickle"

            # We prepend the specified directory name, if given.
            fname = self._within_dir(fname)

            # Here, the pickle file is created and the data is saved.
            with open(fname, "wb") as f:
                pickle.dump(data, f)

            # Store the most recently saved generation number
            self._last_generation = searcher.step_count

            # Store the name of the last pickle file
            self._last_file_name = fname

            # Report to the user that the save was successful
            if self._verbose:
                print("Saved to", fname)

            # Return the file name
            return fname
        else:
            return None

    @property
    def last_generation(self) -> Optional[int]:
        """
        Get the last generation for which a pickle file was created.
        If no pickle file is created yet, the result will be None.
        """
        return self._last_generation

    @property
    def last_file_name(self) -> Optional[str]:
        """
        Get the name of the last pickle file.
        If no pickle file is created yet, the result will be None.
        """
        return self._last_file_name

    def unpickle_last_file(self) -> dict:
        """
        Unpickle the most recently made pickle file.
        The file itself will not be modified.
        Its contents will be returned.
        """
        with open(self._last_file_name, "rb") as f:
            return pickle.load(f)


class ScalarLogger(Logger):
    def _filter(self, status: dict) -> dict:
        short_status = {}
        for k, v in status.items():
            if (not is_sequence(v)) and (not isinstance(v, Solution)):
                short_status[k] = v
        return short_status


class StdOutLogger(ScalarLogger):
    """A logger which prints the status into the screen."""

    def __init__(
        self,
        searcher: SearchAlgorithm,
        *,
        interval: int = 1,
        after_first_step: bool = False,
        leading_keys: Iterable[str] = ("iter",),
    ):
        """`__init__(...)`: Initialize the StdOutLogger.

        Args:
            searcher: The evolutionary algorithm instance whose progress
                is to be logged.
            interval: Expected as an integer n.
                Logging is to be done at every n iterations.
            after_first_step: Expected as a boolean.
                Meaningful only if interval is set as an integer greater
                than 1. Let us suppose that interval is set as 10.
                If after_first_step is False (which is the default),
                then the logging will be done at steps 10, 20, 30, and so on.
                On the other hand, if after_first_step is True,
                then the logging will be done at steps 1, 11, 21, 31, and so
                on.
            leading_keys: A sequence of strings where each string is a status
                key. When printing the status, these keys will be shown first.
        """
        super().__init__(searcher, interval=interval, after_first_step=after_first_step)
        self._leading_keys = list(leading_keys)
        self._leading_keys_set = set(self._leading_keys)

    def _log(self, status: dict):
        max_key_length = max([len(str(k)) for k in status.keys()])

        def report(k, v):
            nonlocal max_key_length
            print(str(k).rjust(max_key_length), ":", v)

        for k in self._leading_keys:
            if k in status:
                v = status[k]
                report(k, v)
        for k, v in status.items():
            if k not in self._leading_keys_set:
                report(k, v)
        print()


if pandas is not None:

    class PandasLogger(ScalarLogger):
        """A logger which collects status information and
        generates a pandas.DataFrame at the end.
        """

        def __init__(self, searcher: SearchAlgorithm, *, interval: int = 1, after_first_step: bool = False):
            """`__init__(...)`: Initialize the PandasLogger.

            Args:
                searcher: The evolutionary algorithm instance whose progress
                    is to be logged.
                interval: Expected as an integer n.
                    Logging is to be done at every n iterations.
                after_first_step: Expected as a boolean.
                    Meaningful only if interval is set as an integer greater
                    than 1. Let us suppose that interval is set as 10.
                    If after_first_step is False (which is the default),
                    then the logging will be done at steps 10, 20, 30, and
                    so on. On the other hand, if after_first_step is True,
                    then the logging will be done at steps 1, 11, 21, 31, and
                    so on.
            """
            super().__init__(searcher, interval=interval, after_first_step=after_first_step)
            self._data = []

        def _log(self, status: dict):
            self._data.append(deepcopy(status))

        def to_dataframe(self, *, index: Optional[str] = "iter") -> pandas.DataFrame:
            """Generate a pandas.DataFrame from the collected
            status information.

            Args:
                index: The column to be set as the index.
                    If passed as None, then no index will be set.
                    The default is "iter".
            """
            result = pandas.DataFrame(self._data)
            if index is not None:
                result.set_index(index, inplace=True)
            return result


if sacred is not None:

    ExpOrRun = Union[sacred.Experiment, sacred.run.Run]

    class SacredLogger(ScalarLogger):
        """A logger which stores the status via the Run object of sacred."""

        def __init__(
            self,
            searcher: SearchAlgorithm,
            run: ExpOrRun,
            result: Optional[str] = None,
            *,
            interval: int = 1,
            after_first_step: bool = False,
        ):
            """`__init__(...)`: Initialize the SacredLogger.

            Args:
                searcher: The evolutionary algorithm instance whose progress
                    is to be logged.
                run: An instance of `sacred.run.Run` or `sacred.Experiment`,
                    using which the progress will be logged.
                result: The key in the status dictionary whose associated
                    value will be registered as the current result
                    of the experiment.
                    If left as None, no result will be registered.
                interval: Expected as an integer n.
                    Logging is to be done at every n iterations.
                after_first_step: Expected as a boolean.
                    Meaningful only if interval is set as an integer greater
                    than 1. Let us suppose that interval is set as 10.
                    If after_first_step is False (which is the default),
                    then the logging will be done at steps 10, 20, 30, and
                    so on. On the other hand, if after_first_step is True,
                    then the logging will be done at steps 1, 11, 21, 31,
                    and so on.
            """
            super().__init__(searcher, interval=interval, after_first_step=after_first_step)
            self._result = result
            self._run = run

        def _log(self, status: dict):
            for k, v in status.items():
                self._run.log_scalar(k, v)
            if self._result is not None:
                self._run.result = status[self._result]


if mlflow is not None:
    MlflowID = Union[str, bytes, int]

    class MlflowLogger(ScalarLogger):
        """A logger which stores the status via Mlflow."""

        def __init__(
            self,
            searcher: SearchAlgorithm,
            client: Optional[mlflow.tracking.MlflowClient] = None,
            run: Union[mlflow.entities.Run, Optional[MlflowID]] = None,
            *,
            interval: int = 1,
            after_first_step: bool = False,
        ):
            """`__init__(...)`: Initialize the MlflowLogger.

            Args:
                searcher: The evolutionary algorithm instance whose progress
                    is to be logged.
                client: The MlflowClient object whose log_metric() method
                    will be used for logging. This can be passed as None,
                    in which case mlflow.log_metrics() will be used instead.
                    Please note that, if a client is provided, the `run`
                    argument is required as well.
                run: Expected only if a client is provided.
                    This is the mlflow Run object (an instance of
                    mlflow.entities.Run), or the ID of the mlflow run.
                interval: Expected as an integer n.
                    Logging is to be done at every n iterations.
                after_first_step: Expected as a boolean.
                    Meaningful only if interval is set as an integer greater
                    than 1. Let us suppose that interval is set as 10.
                    If after_first_step is False (which is the default),
                    then the logging will be done at steps 10, 20, 30, and
                    so on. On the other hand, if after_first_step is True,
                    then the logging will be done at steps 1, 11, 21, 31,
                    and so on.
            """

            super().__init__(searcher, interval=interval, after_first_step=after_first_step)

            self._client = client
            self._run_id: Optional[MlflowID] = None

            if self._client is None:
                if run is not None:
                    raise ValueError("Received `run`, but `client` is missing")
            else:
                if run is None:
                    raise ValueError("Received `client`, but `run` is missing")
                if isinstance(run, mlflow.entities.Run):
                    self._run_id = run.info.run_id
                else:
                    self._run_id = run

        def _log(self, status: dict):
            if self._client is None:
                mlflow.log_metrics(status, step=self._steps_count)
            else:
                for k, v in status.items():
                    self._client.log_metric(self._run_id, k, v, step=self._steps_count)


if neptune is not None:

    class NeptuneLogger(ScalarLogger):
        """A logger which stores the status via neptune."""

        def __init__(
            self,
            searcher: SearchAlgorithm,
            run,
            *,
            interval: int = 1,
            after_first_step: bool = False,
            group: Optional[str] = None,
        ):
            """`__init__(...)`: Initialize the NeptuneLogger.

            Args:
                searcher: The evolutionary algorithm instance whose progress
                    is to be logged.
                run: A `neptune.new.run.Run` instance using which the status
                    will be logged.
                interval: Expected as an integer n.
                    Logging is to be done at every n iterations.
                after_first_step: Expected as a boolean.
                    Meaningful only if interval is set as an integer greater
                    than 1. Let us suppose that interval is set as 10.
                    If after_first_step is False (which is the default),
                    then the logging will be done at steps 10, 20, 30, and so on.
                    On the other hand, if after_first_step is True,
                    then the logging will be done at steps 1, 11, 21, 31, and so
                    on.
                group: Into which group will the metrics be stored.
                    For example, if the status keys to be logged are "score" and
                    "elapsed", and `group` is set as "training", then the metrics
                    will be sent to neptune with the keys "training/score" and
                    "training/elapsed". `group` can also be left as None,
                    in which case the status will be sent to neptune with the
                    key names unchanged.
            """
            super().__init__(searcher, interval=interval, after_first_step=after_first_step)
            self._run = run
            self._group = group

        def _log(self, status: dict):
            for k, v in status.items():
                target_key = k if self._group is None else self._group + "/" + k
                self._run[target_key].log(v)



# Copyright 2022 NNAISENSE SA
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from collections import OrderedDict
from copy import deepcopy
from numbers import Number
from typing import Any, Optional

import numpy as np
import torch


def deep_clone(  # noqa: C901
    x: Any,
    *,
    otherwise_deepcopy: bool = False,
    otherwise_return: bool = False,
    otherwise_fail: bool = False,
    memo: Optional[dict] = None,
) -> Any:
    """
    A recursive cloning function similar to the standard `deepcopy`.

    The difference between `deep_clone(...)` and `deepcopy(...)` is that
    `deep_clone(...)`, while recursively traversing, will run the `.clone()`
    method on the PyTorch tensors it encounters, so that the cloned tensors
    are forcefully detached from their storages (instead of cloning those
    storages as well).

    At the moment of writing this documentation, the current behavior of
    PyTorch tensors upon being deep-copied is to clone themselves AND their
    storages. Therefore, if a PyTorch tensor is a slice of a large tensor
    (which has a large storage), then the large storage will also be
    deep-copied, and the newly made clone of the tensor will point to a newly
    made large storage. One might instead prefer to clone tensors in such a
    way that the newly made tensor points to a newly made storage that
    contains just enough data for the tensor (with the unused data being
    dropped). When such a behavior is desired, one can use this
    `deep_clone(...)` function.

    Upon encountering a read-only and/or immutable data, this function will
    NOT modify the read-only behavior. For example, the deep-clone of a
    ReadOnlyTensor is still a ReadOnlyTensor, and the deep-clone of a
    read-only numpy array is still a read-only numpy array. Note that this
    behavior is different than the `clone()` method of a ReadOnlyTensor
    and the `copy()` method of a numpy array. The reason for this
    protective behavior is that since this is a deep-cloning operation,
    the encountered tensors and/or arrays might be the components of the root
    object, and changing their read-only attributes might affect the integrity
    of this root object.

    The `deep_clone(...)` function needs to know what to do when an object
    of unrecognized type is encountered. Therefore, the user is expected to
    set one of these arguments as True (and leave the others as False):
    `otherwise_deepcopy`, `otherwise_return`, `otherwise_fail`.

    Args:
        x: The object which will be deep-cloned. This object can be a standard
            Python container (i.e. list, tuple, dict, set), an instance of
            Problem, Solution, SolutionBatch, ObjectArray, ImmutableContainer,
            Clonable, and also any other type of object if either the argument
            `otherwise_deepcopy` or the argument `otherwise_return` is set as
            True.
        otherwise_deepcopy: Setting this as True means that, when an
            unrecognized object is encountered, that object will be
            deep-copied. To handle shared and cyclic-referencing objects,
            the `deep_clone(...)` function stores its own memo dictionary.
            When the control is given to the standard `deepcopy(...)`
            function, the memo dictionary of `deep_clone(...)` will be passed
            to `deepcopy`.
        otherwise_return: Setting this as True means that, when an
            unrecognized object is encountered, that object itself will be
            returned (i.e. will be a part of the created clone).
        otherwise_fail: Setting this as True means that, when an unrecognized
            object is encountered, a TypeError will be raised.
        memo: Optionally a dictionary. In most scenarios, when this function
            is called from outside, this is expected to be left as None.
    Returns:
        The newly made clone of the original object.
    """
    from .objectarray import ObjectArray
    from .readonlytensor import ReadOnlyTensor

    if memo is None:
        # If a memo dictionary was not given, make a new one now.
        memo = {}

    # Get the id of the object being cloned.
    x_id = id(x)

    if x_id in memo:
        # If the id of the object being cloned is already in the memo dictionary, then this object was previously
        # cloned. We just return that clone.
        return memo[x_id]

    # Count how many of the arguments `otherwise_deepcopy`, `otherwise_return`, and `otherwise_fail` was set as True.
    # In this context, we call these arguments as fallback behaviors.
    fallback_behaviors = (otherwise_deepcopy, otherwise_return, otherwise_fail)
    enabled_behavior_count = sum(1 for behavior in fallback_behaviors if behavior)

    if enabled_behavior_count == 0:
        # If none of the fallback behaviors was enabled, then we raise an error.
        raise ValueError(
            "The action to take with objects of unrecognized types is not known because"
            " none of these arguments was set as True: `otherwise_deepcopy`, `otherwise_return`, `otherwise_fail`."
            " Please set one of these arguments as True."
        )
    elif enabled_behavior_count == 1:
        # If one of the fallback behaviors was enabled, then we received our expected input. We do nothing here.
        pass
    else:
        # If the number of enabled fallback behaviors is an unexpected value. then we raise an error.
        raise ValueError(
            f"The following arguments were received, which is conflicting: otherwise_deepcopy={otherwise_deepcopy},"
            f" otherwise_return={otherwise_return}, otherwise_fail={otherwise_fail}."
            f" Please set exactly one of these arguments as True and leave the others as False."
        )

    # This inner function specifies how the deep_clone function should call itself.
    def call_self(obj: Any) -> Any:
        return deep_clone(
            obj,
            otherwise_deepcopy=otherwise_deepcopy,
            otherwise_return=otherwise_return,
            otherwise_fail=otherwise_fail,
            memo=memo,
        )

    # Below, we handle the cloning behaviors case by case.
    if (x is None) or (x is NotImplemented) or (x is Ellipsis):
        result = deepcopy(x)
    elif isinstance(x, (Number, str, bytes, bytearray)):
        result = deepcopy(x, memo=memo)
    elif isinstance(x, np.ndarray):
        result = x.copy()
        result.flags["WRITEABLE"] = x.flags["WRITEABLE"]
    elif isinstance(x, (ObjectArray, ReadOnlyClonable)):
        result = x.clone(preserve_read_only=True, memo=memo)
    elif isinstance(x, ReadOnlyTensor):
        result = x.clone(preserve_read_only=True)
    elif isinstance(x, torch.Tensor):
        result = x.clone()
    elif isinstance(x, Clonable):
        result = x.clone(memo=memo)
    elif isinstance(x, (dict, OrderedDict)):
        result = type(x)()
        memo[x_id] = result
        for k, v in x.items():
            result[call_self(k)] = call_self(v)
    elif isinstance(x, list):
        result = type(x)()
        memo[x_id] = result
        for item in x:
            result.append(call_self(item))
    elif isinstance(x, set):
        result = type(x)()
        memo[x_id] = result
        for item in x:
            result.add(call_self(item))
    elif isinstance(x, tuple):
        result = []
        memo[x_id] = result
        for item in x:
            result.append(call_self(item))
        if hasattr(x, "_fields"):
            result = type(x)(*result)
        else:
            result = type(x)(result)
        memo[x_id] = result
    else:
        # If the object is not recognized, we use the fallback behavior.
        if otherwise_deepcopy:
            result = deepcopy(x, memo=memo)
        elif otherwise_return:
            result = x
        elif otherwise_fail:
            raise TypeError(f"Do not know how to clone {repr(x)} (of type {type(x)}).")
        else:
            raise RuntimeError("The function `deep_clone` reached an unexpected state. This might be a bug.")

    if (x_id not in memo) and (result is not x):
        # If the newly made clone is still not in the memo dictionary AND the "clone" is not just a reference to the
        # original object, we make sure that it is in the memo dictionary.
        memo[x_id] = result

    # Finally, the result is returned.
    return result


class Clonable:
    """
    A base class allowing inheriting classes define how they should be cloned.

    Any class inheriting from Clonable gains these behaviors:
    (i) A new method named `.clone()` becomes available;
    (ii) `__deepcopy__` and `__copy__` work as aliases for `.clone()`;
    (iii) A new method, `_get_cloned_state(self, *, memo: dict)` is now
    defined and needs to be implemented by the inheriting class.

    The method `_get_cloned_state(...)` expects a dictionary named `memo`,
    which maps from the ids of already cloned objects to their clones.
    If `_get_cloned_state(...)` is to use `deep_clone(...)` or `deepcopy(...)`
    within itself, this `memo` dictionary can be passed to these functions.
    The return value of `_get_cloned_state(...)` is a dictionary, which will
    be used as the `__dict__` of the newly made clone.
    """

    def _get_cloned_state(self, *, memo: dict) -> dict:
        raise NotImplementedError

    def clone(self, *, memo: Optional[dict] = None) -> "Clonable":
        """
        Get a clone of this object.

        Args:
            memo: Optionally a dictionary which maps from the ids of the
                already cloned objects to their clones. In most scenarios,
                when this method is called from outside, this can be left
                as None.
        Returns:
            The clone of the object.
        """
        if memo is None:
            memo = {}

        self_id = id(self)
        if self_id in memo:
            return memo[self_id]

        new_object = object.__new__(type(self))
        memo[id(self)] = new_object

        new_object.__dict__.update(self._get_cloned_state(memo=memo))
        return new_object

    def __copy__(self) -> "Clonable":
        return self.clone()

    def __deepcopy__(self, memo: Optional[dict]):
        if memo is None:
            memo = {}
        return self.clone(memo=memo)


class Serializable(Clonable):
    """
    Base class allowing the inheriting classes become Clonable and picklable.

    Any class inheriting from `Serializable` becomes `Clonable` (since
    `Serializable` is a subclass of `Clonable`) and therefore is expected to
    define its own `_get_cloned_state(...)` (see the documentation of the
    class `Clonable` for details).

    A `Serializable` class gains a behavior for its `__getstate__`. In this
    already defined and implemented `__getstate__` method, the resulting
    dictionary of `_get_cloned_state(...)` is used as the state dictionary.
    Therefore, for `Serializable` objects, the behavior defined in their
    `_get_cloned_state(...)` methods affect how they are pickled.

    Classes inheriting from `Serializable` are `evotorch.Problem`,
    `evotorch.Solution`, `evotorch.SolutionBatch`, and
    `evotorch.distributions.Distribution`. In their `_get_cloned_state(...)`
    implementations, these classes use `deep_clone(...)` on themselves to make
    sure that their contained PyTorch tensors are copied using the `.clone()`
    method, ensuring that those tensors are detached from their old storages
    during the cloning operation. Thanks to being `Serializable`, their
    contained tensors are detached from their old storages both at the moment
    of copying/cloning AND at the moment of pickling.
    """

    def __getstate__(self) -> dict:
        memo = {id(self): self}
        return self._get_cloned_state(memo=memo)


class ReadOnlyClonable(Clonable):
    """
    Clonability base class for read-only and/or immutable objects.

    This is a base class specialized for the immutable containers of EvoTorch.
    These immutable containers have two behaviors for cloning:
    one where the read-only attribute is preserved and one where a mutable
    clone is created.

    Upon being copied or deep-copied (using the standard Python functions),
    the newly made clones are also read-only. However, when copied using the
    `clone(...)` method, the newly made clone is mutable by default
    (unless the `clone(...)` method was used with `preserve_read_only=True`).
    This default behavior of the `clone(...)` method was inspired by the
    `copy()` method of numpy arrays (the inspiration being that the `.copy()`
    of a read-only numpy array will not be read-only anymore).

    Subclasses of `evotorch.immutable.ImmutableContainer` inherit from
    `ReadOnlyClonable`.
    """

    def _get_mutable_clone(self, *, memo: dict) -> Any:
        raise NotImplementedError

    def clone(self, *, memo: Optional[dict] = None, preserve_read_only: bool = False) -> Any:
        """
        Get a clone of this read-only object.

        Args:
            memo: Optionally a dictionary which maps from the ids of the
                already cloned objects to their clones. In most scenarios,
                when this method is called from outside, this can be left
                as None.
            preserve_read_only: Whether or not to preserve the read-only
                behavior in the clone.
        Returns:
            The clone of the object.
        """
        if memo is None:
            memo = {}
        if preserve_read_only:
            return super().clone(memo=memo)
        else:
            return self._get_mutable_clone(memo=memo)

    def __copy__(self) -> Any:
        return self.clone(preserve_read_only=True)

    def __deepcopy__(self, memo: Optional[dict]) -> Any:
        if memo is None:
            memo = {}
        return self.clone(memo=memo, preserve_read_only=True)



# Copyright 2022 NNAISENSE SA
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import pytest
import torch

from evotorch.testing import assert_allclose
from evotorch.tools import cast_tensors_in_container, device_of_container, dtype_of_container


@pytest.mark.parametrize("dtype", (torch.float32, torch.float64, torch.int32, torch.int64))
def test_dtype_of_container(dtype):
    x = {
        "a": torch.tensor([1, 2, 3], dtype=dtype),
        "b": [1, 2],
        "c": (
            torch.tensor([10, 20], dtype=dtype),
            "some string",
        ),
    }

    assert dtype_of_container(x) == dtype


@pytest.mark.parametrize("dtype", (torch.float32, torch.float64, torch.int32, torch.int64))
def test_dtype_of_container_cyclic(dtype):
    x = {
        "a": torch.tensor([1, 2, 3], dtype=dtype),
        "b": [1, 2],
        "c": (
            torch.tensor([10, 20], dtype=dtype),
            "some string",
        ),
    }

    x["b"].append(x)
    x["d"] = x["a"]

    assert dtype_of_container(x) == dtype


def test_dtype_of_container_none():
    x = {
        "a": [1, 2, 3],
        "b": [1, 2],
        "c": (
            [10, 20],
            "some string",
        ),
    }

    x["b"].append(x)
    x["d"] = x["a"]

    assert dtype_of_container(x) is None


def test_device_of_container():
    device = torch.device("cpu")

    x = {
        "a": torch.tensor([1, 2, 3], device=device),
        "b": [1, 2],
        "c": (
            torch.tensor([10, 20], device=device),
            "some string",
        ),
    }

    assert device_of_container(x) == device


def test_device_of_container_cyclic():
    device = torch.device("cpu")

    x = {
        "a": torch.tensor([1, 2, 3], device=device),
        "b": [1, 2],
        "c": (
            torch.tensor([10, 20], device=device),
            "some string",
        ),
    }

    x["b"].append(x)
    x["d"] = x["a"]

    assert device_of_container(x) == device


def test_device_of_container_none():
    x = {
        "a": [1, 2, 3],
        "b": [1, 2],
        "c": (
            [10, 20],
            "some string",
        ),
    }

    x["b"].append(x)
    x["d"] = x["a"]

    assert device_of_container(x) is None


def _assert_equals_but_not_same(a, b, atol):
    assert a is not b
    assert_allclose(a, b, atol=atol)


@pytest.mark.parametrize(
    "from_dtype,to_dtype",
    (
        (torch.float32, torch.float64),
        (torch.float64, torch.float32),
    ),
)
def test_cast_tensors_in_container(from_dtype, to_dtype):
    tolerance = 1e-4

    x = {
        "a": torch.randn(10, dtype=from_dtype),
        "b": [1, 2],
        "c": (
            torch.randn(5, dtype=from_dtype),
            "some string",
        ),
    }

    # Make another element point to an existing tensor
    x["d"] = x["c"][0]

    # Introduce a cyclic reference
    x["e"] = x

    # Cast all the tensors in x
    y = cast_tensors_in_container(x, dtype=to_dtype)

    # Assert that the resulting dictionary has the same keys with the original
    assert set(y.keys()) == set(x.keys())

    # Assert that the resulting dictionary is NOT the same dictionary with the original
    assert y is not x

    # Assert that the cloning was done correctly
    assert isinstance(y["a"], torch.Tensor)
    _assert_equals_but_not_same(y["a"], x["a"], atol=tolerance)

    assert isinstance(y["b"], list)
    assert y["b"] is not x["b"]
    assert len(y["b"]) == len(x["b"])
    assert y["b"] == x["b"]

    assert isinstance(y["c"], tuple)
    assert y["c"] is not x["c"]
    assert len(y["c"]) == len(x["c"])
    _assert_equals_but_not_same(y["c"][0], x["c"][0], atol=tolerance)
    assert y["c"][1] == x["c"][1]

    assert y["d"] is not x["d"]
    assert y["d"] is y["c"][0]

    assert y["e"] is not x
    assert y["e"] is y



# Copyright 2022 NNAISENSE SA
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Utilities for reading and for writing neural network parameters"""

from typing import Optional, Union

import torch
from torch import nn

from ...tools.misc import Device


@torch.no_grad()
def fill_parameters(net: nn.Module, vector: torch.Tensor):
    """Fill the parameters of a torch module (net) from a vector.

    No gradient information is kept.

    The vector's length must be exactly the same with the number
    of parameters of the PyTorch module.

    Args:
        net: The torch module whose parameter values will be filled.
        vector: A 1-D torch tensor which stores the parameter values.
    """
    address = 0
    for p in net.parameters():
        d = p.data.view(-1)
        n = len(d)
        d[:] = torch.as_tensor(vector[address : address + n], device=d.device)
        address += n

    if address != len(vector):
        raise IndexError("The parameter vector is larger than expected")


@torch.no_grad()
def parameter_vector(net: nn.Module, *, device: Optional[Device] = None) -> torch.Tensor:
    """Get all the parameters of a torch module (net) into a vector

    No gradient information is kept.

    Args:
        net: The torch module whose parameters will be extracted.
        device: The device in which the parameter vector will be constructed.
            If the network has parameter across multiple devices,
            you can specify this argument so that concatenation of all the
            parameters will be successful.
    Returns:
        The parameters of the module in a 1-D tensor.
    """
    dev_kwarg = {} if device is None else {"device": device}

    all_vectors = []
    for p in net.parameters():
        all_vectors.append(torch.as_tensor(p.data.view(-1), **dev_kwarg))

    return torch.cat(all_vectors)


def count_parameters(net: nn.Module) -> int:
    """
    Get the number of parameters the network.

    Args:
        net: The torch module whose parameters will be counted.
    Returns:
        The number of parameters, as an integer.
    """

    count = 0

    for p in net.parameters():
        count += p.numel()

    return count


def device_of_module(m: nn.Module, default: Optional[Union[str, torch.device]] = None) -> torch.device:
    """
    Get the device in which the module exists.

    This function looks at the first parameter of the module, and returns
    its device. This function is not meant to be used on modules whose
    parameters exist on different devices.

    Args:
        m: The module whose device is being queried.
        default: The fallback device to return if the module has no
            parameters. If this is left as None, the fallback device
            is assumed to be "cpu".
    Returns:
        The device of the module, determined from its first parameter.
    """
    if default is None:
        default = torch.device("cpu")

    device = default

    for p in m.parameters():
        device = p.device
        break

    return device



# Copyright 2022 NNAISENSE SA
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from typing import Any, Iterable, Mapping


class RecursivePrintable:
    """
    A base class for making a class printable.

    This base class considers custom container types which can recursively
    contain themselves (even in a cyclic manner). Classes inheriting from
    `RecursivePrintable` will gain a new ready-to-use method named
    `to_string(...)`. This `to_string(...)` method, upon being called,
    checks if the current class is an Iterable or a Mapping, and prints
    the representation accordingly, with a recursion limit to avoid
    `RecursionError`. The methods `__str__(...)` and `__repr__(...)`
    are also defined as aliases of this `to_string` method.
    """

    def to_string(self, *, max_depth: int = 10) -> str:
        if max_depth <= 0:
            return "<...>"

        def item_repr(x: Any) -> str:
            if isinstance(x, RecursivePrintable):
                return x.to_string(max_depth=(max_depth - 1))
            else:
                return repr(x)

        result = []

        def puts(*x: Any):
            for item_of_x in x:
                result.append(str(item_of_x))

        clsname = type(self).__name__
        first_one = True

        if isinstance(self, Mapping):
            puts(clsname, "({")
            for k, v in self.items():
                if first_one:
                    first_one = False
                else:
                    puts(", ")
                puts(item_repr(k), ": ", item_repr(v))
            puts("})")
        elif isinstance(self, Iterable):
            puts(clsname, "([")
            for v in self:
                if first_one:
                    first_one = False
                else:
                    puts(", ")
                puts(item_repr(v))
            puts("])")
        else:
            raise NotImplementedError

        return "".join(result)

    def __str__(self) -> str:
        return self.to_string()

    def __repr__(self) -> str:
        return self.to_string()



# Copyright 2022 NNAISENSE SA
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from ..core import Problem


class BaseNEProblem(Problem):
    """
    This is the base class for all neuro-evolution problems.

    Currently, this class does not offer any additional functionality.
    Its purpose is to collect all neuro-evolution problems under the same
    branch of inheritance.
    """

    pass



# Copyright 2022 NNAISENSE SA
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import torch

from evotorch import Problem
from evotorch.algorithms import SteadyStateGA
from evotorch.logging import StdOutLogger
from evotorch.operators import GaussianMutation, SimulatedBinaryCrossOver


# Kursawe function with two conflicting objectives
# Non-vectorized (evaluates a single solution vector) for illustrative purposes
def kursawe(x: torch.Tensor) -> torch.Tensor:
    f1 = torch.sum(-10 * torch.exp(-0.2 * torch.sqrt(x[0:2] ** 2.0 + x[1:3] ** 2.0)), dim=-1)
    f2 = torch.sum((torch.abs(x) ** 0.8) + (5 * torch.sin(x**3)), dim=-1)
    fitnesses = torch.stack([f1, f2], dim=-1)
    return fitnesses


problem = Problem(
    # Two objectives, both minimization
    ["min", "min"],
    kursawe,
    initial_bounds=(-5.0, 5.0),
    solution_length=3,
    vectorized=False,
    # Create as many parallel Ray actors as there are available CPUs
    num_actors="max",
)

# Run a GA similar to NSGA-II for 100 steps and log results to standard output every 10 steps
searcher = SteadyStateGA(problem, popsize=200)
searcher.use(SimulatedBinaryCrossOver(problem, tournament_size=4, cross_over_rate=1.0, eta=8))
searcher.use(GaussianMutation(problem, stdev=0.03))
logger = StdOutLogger(searcher, interval=10)

searcher.run(100)
print("Final status:\n", searcher.status)



# Copyright 2022 NNAISENSE SA
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from collections import OrderedDict
from collections.abc import Iterable, Mapping, Sequence, Set
from numbers import Number
from typing import Any, Optional, Union

import numpy as np
import torch

from .cloning import ReadOnlyClonable, deep_clone
from .recursiveprintable import RecursivePrintable


def _is_basic_data(x: Any) -> bool:
    return (x is None) or isinstance(x, (str, Number))


def _numpy_array_stores_objects(x: np.ndarray) -> bool:
    return x.dtype == np.dtype(object)


def _is_numpy_array_immutable(x: np.ndarray) -> bool:
    return (not _numpy_array_stores_objects(x)) and (not x.flags["WRITEABLE"])


def is_immutable_container_or_tensor(x: Any) -> bool:
    from .objectarray import ObjectArray
    from .readonlytensor import ReadOnlyTensor

    return (
        isinstance(x, (ImmutableContainer, ReadOnlyTensor))
        or (isinstance(x, np.ndarray) and _is_numpy_array_immutable(x))
        or (isinstance(x, ObjectArray) and x.is_read_only)
    )


def as_immutable(x: Any, *, memo: Optional[dict] = None) -> Any:
    from .objectarray import ObjectArray
    from .readonlytensor import ReadOnlyTensor

    if memo is None:
        memo = {}

    x_id = id(x)
    if x_id in memo:
        return memo[x_id]

    put_into_memo = True

    if is_immutable_container_or_tensor(x) or _is_basic_data(x):
        put_into_memo = False
        result = x
    elif isinstance(x, torch.Tensor):
        result = x.clone().as_subclass(ReadOnlyTensor)
    elif isinstance(x, ObjectArray):
        result = x.clone().get_read_only_view()
    elif isinstance(x, np.ndarray):
        if _numpy_array_stores_objects(x):
            result = ObjectArray(len(x))
            for i, element in enumerate(x):
                result[i] = element
            result = result.get_read_only_view()
        else:
            result = x.copy()
            result.flags["WRITEABLE"] = False
            result = result
    elif isinstance(x, Mapping):
        result = ImmutableDict(x, memo)
    elif isinstance(x, set):
        result = ImmutableSet(x, memo=memo)
    elif isinstance(x, Iterable):
        result = ImmutableList(x, memo=memo)
    else:
        raise TypeError(
            f"as_immutable(...) encountered an object of unsupported type."
            f" Encountered object: {repr(x)}."
            f" Type of the encountered object: {repr(type(x))}."
        )

    if put_into_memo:
        memo[x_id] = result

    return result


def mutable_copy(x: Any, *, memo: Optional[dict] = None) -> Any:
    from .objectarray import ObjectArray
    from .readonlytensor import ReadOnlyTensor

    if memo is None:
        memo = {}

    x_id = id(x)
    if x_id in memo:
        return memo[x_id]

    if _is_basic_data(x):
        result = x
    elif isinstance(x, (ReadOnlyTensor, ObjectArray)):
        result = x.clone(preserve_read_only=False)
    elif isinstance(x, np.ndarray):
        result = x.copy()
    elif isinstance(x, torch.Tensor):
        result = x.clone()
    elif isinstance(x, ImmutableContainer):
        result = x.clone(memo=memo, preserve_read_only=False)
    else:
        raise TypeError(f"Encountered an object of unrecognized type. The object is {repr(x)}. Its type is {type(x)}")

    if (x_id not in memo) and (result is not x):
        memo[x_id] = result

    return result


class ImmutableContainer(ReadOnlyClonable, RecursivePrintable):
    def _get_cloned_state(self, *, memo: dict) -> dict:
        return deep_clone(self.__dict__, otherwise_deepcopy=True, memo=memo)

    def __setstate__(self, state: dict):
        self.__dict__.update(state)

        # After pickling and unpickling, numpy arrays become mutable.
        # Since we are dealing with immutable containers here, we need to forcefully make all numpy arrays read-only.
        if isinstance(self, Mapping):
            all_values = self.values()
        elif isinstance(self, Sequence):
            all_values = self
        else:
            raise NotImplementedError

        for v in all_values:
            if isinstance(v, np.ndarray):
                v.flags["WRITEABLE"] = False


class ImmutableSequence(ImmutableContainer, Sequence):
    def __init__(self, x: Iterable, *, memo: Optional[dict] = None):
        if memo is None:
            memo = {}
        memo[id(x)] = self
        self.__data: list = [as_immutable(item, memo=memo) for item in x]

    def __len__(self) -> int:
        return len(self.__data)

    def __getitem__(self, i):
        return self.__data[i]

    def __add__(self, other: Iterable) -> "ImmutableSequence":
        cls = type(self)
        return cls(list(self) + list(other))

    def __radd__(self, other: Iterable) -> "ImmutableSequence":
        cls = type(self)
        return cls(list(other) + list(self))

    def __eq__(self, other: Iterable) -> bool:
        self_length = len(self)
        self_last = self_length - 1
        other_count = 0
        for i, other_item in enumerate(other):
            other_count += 1
            if i > self_last:
                return False
            if self[i] != other_item:
                return False
        return other_count == self_length

    def __ne__(self, other: Iterable) -> bool:
        return not self.__eq__(other)


class ImmutableList(ImmutableSequence):
    def _get_mutable_clone(self, *, memo: dict) -> list:
        result = []
        memo[id(self)] = result
        for x in self:
            result.append(mutable_copy(x, memo=memo))
        return result


class ImmutableSet(ImmutableContainer, Set):
    def __init__(self, x: Iterable, *, memo: Optional[dict] = None):
        if memo is None:
            memo = {}
        memo[id(x)] = self
        self.__data: set = set([as_immutable(item, memo=memo) for item in x])

    def __contains__(self, x: Any) -> bool:
        return as_immutable(x) in self.__data

    def __iter__(self):
        for item in self.__data:
            yield item

    def __len__(self) -> int:
        return len(self.__data)

    def _get_mutable_clone(self, *, memo: dict) -> set:
        result = set()
        memo[id(self)] = result
        for x in self:
            result.add(mutable_copy(x, memo=memo))
        return result


def _acceptable_key(x: Any) -> Union[int, str]:
    from .misc import is_integer

    if is_integer(x):
        return int(x)
    elif isinstance(x, str):
        return x
    elif isinstance(x, (tuple, ImmutableSequence)):
        return tuple([_acceptable_key(k) for k in x])
    else:
        raise TypeError(
            f"The object to be used as a key within an `ImmutableDict` must"
            f" be of one of these types: `int`, `str`, `tuple` (where the"
            f" tuple can only contain `int`, `str` and/or `tuple` instances)."
            f" The received key candidate does not seem to satisfy this"
            f" constraint."
            f" The unnaccepted key candidate is: {repr(x)}."
            f" The type of the unnaccepted key is: {repr(type(x))}."
        )


class ImmutableDict(ImmutableContainer, Mapping):
    def __init__(self, *args, **kwargs):
        if len(args) == 1:
            x = args[0]
            memo = {}
        elif len(args) == 2:
            x, memo = args
            if memo is None:
                memo = {}
        else:
            raise TypeError("Wrong number of positional arguments. Expected 1 or 2 arguments.")

        memo[id(x)] = self

        self.__data: OrderedDict = OrderedDict()
        iterator = x.items() if isinstance(x, Mapping) else x
        for k, v in iterator:
            self.__data[_acceptable_key(k)] = as_immutable(v, memo=memo)
        for k, v in kwargs.items():
            self.__data[str(k)] = as_immutable(v, memo=memo)

    def __getitem__(self, k):
        k = _acceptable_key(k)
        return self.__data[k]

    def __iter__(self):
        for k in self.__data.keys():
            yield k

    def __len__(self) -> int:
        return len(self.__data)

    def _get_mutable_clone(self, *, memo: dict) -> dict:
        result = {}
        memo[id(self)] = result
        for k, v in self.items():
            k = mutable_copy(k, memo=memo)
            v = mutable_copy(v, memo=memo)
            result[k] = v
        return result



# Copyright 2022 NNAISENSE SA
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Utility classes and functions for neural networks"""

__all__ = (
    "count_parameters",
    "fill_parameters",
    "parameter_vector",
    "device_of_module",
    "layers",
    "misc",
    "rl",
    "vecrl",
    "multilayered",
    "MultiLayered",
    "statefulmodule",
    "StatefulModule",
    "parser",
    "str_to_net",
    "Policy",
    "RunningStat",
    "RunningNorm",
    "functional",
    "ModuleExpectingFlatParameters",
    "make_functional_module",
)

from . import functional, layers, misc, multilayered, parser, rl, statefulmodule, vecrl
from .functional import ModuleExpectingFlatParameters, make_functional_module
from .misc import count_parameters, device_of_module, fill_parameters, parameter_vector
from .multilayered import MultiLayered
from .parser import str_to_net
from .runningnorm import RunningNorm
from .runningstat import RunningStat
from .statefulmodule import StatefulModule
from .vecrl import Policy



# Copyright 2022 NNAISENSE SA
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from copy import deepcopy
from typing import Optional, Union

import numpy as np
import torch
from torch import nn

from .runningnorm import RunningNorm


class RunningStat:
    """
    Tool for efficiently computing the mean and stdev of arrays.
    The arrays themselves are not stored separately,
    instead, they are accumulated.

    This RunningStat is implemented as a wrapper around RunningNorm.
    The difference is that the interface of RunningStat is simplified
    to expect only numpy arrays, and expect only non-vectorized
    observations.
    With this simplified interface, RunningStat is meant to be used
    by GymNE, on classical non-vectorized gym tasks.
    """

    def __init__(self):
        """
        `__init__(...)`: Initialize the RunningStat.
        """
        self._rn: Optional[RunningNorm] = None
        self.reset()

    def reset(self):
        """
        Reset the RunningStat to its initial state.
        """
        self._rn = None

    @property
    def count(self) -> int:
        """
        Get the number of arrays accumulated.
        """
        if self._rn is None:
            return 0
        else:
            return self._rn.count

    @property
    def sum(self) -> np.ndarray:
        """
        Get the sum of all accumulated arrays.
        """
        return self._rn.sum.numpy()

    @property
    def sum_of_squares(self) -> np.ndarray:
        """
        Get the sum of squares of all accumulated arrays.
        """
        return self._rn.sum_of_squares.numpy()

    @property
    def mean(self) -> np.ndarray:
        """
        Get the mean of all accumulated arrays.
        """
        return self._rn.mean.numpy()

    @property
    def stdev(self) -> np.ndarray:
        """
        Get the standard deviation of all accumulated arrays.
        """
        return self._rn.stdev.numpy()

    def update(self, x: Union[np.ndarray, "RunningStat"]):
        """
        Accumulate more data into the RunningStat object.
        If the argument is an array, that array is added
        as one more data element.
        If the argument is another RunningStat instance,
        all the stats accumulated by that RunningStat object
        are added into this RunningStat object.
        """
        if isinstance(x, RunningStat):
            if x.count > 0:
                if self._rn is None:
                    self._rn = deepcopy(x._rn)
                else:
                    self._rn.update(x._rn)
        else:
            if self._rn is None:
                x = np.array(x, dtype="float32")
                self._rn = RunningNorm(shape=x.shape, dtype="float32", device="cpu")
            self._rn.update(x)

    def normalize(self, x: Union[np.ndarray, list]) -> np.ndarray:
        """
        Normalize the array x according to the accumulated stats.
        """
        if self._rn is None:
            return x
        else:
            x = np.array(x, dtype="float32")
            return self._rn.normalize(x)

    def __copy__(self):
        return deepcopy(self)

    def __repr__(self) -> str:
        return f"<{self.__class__.__name__}, count: {self.count}>"

    def to(self, device: Union[str, torch.device]) -> "RunningStat":
        """
        If the target device is cpu, return this RunningStat instance itself.
        A RunningStat object is meant to work with numpy arrays. Therefore,
        any device other than the cpu will trigger an error.

        Args:
            device: The target device. Only cpu is supported.
        Returns:
            The original RunningStat.
        """
        if torch.device(device) == torch.device("cpu"):
            return self
        else:
            raise ValueError(
                f"The received target device is {repr(device)}. However, RunningStat can only work on a cpu."
            )

    def to_layer(self) -> nn.Module:
        """
        Make a PyTorch module which normalizes the its inputs.

        Returns:
            An ObsNormLayer instance.
        """
        return self._rn.to_layer()



# Copyright 2022 NNAISENSE SA
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Optional

import torch
from torch import nn


class MultiLayered(nn.Module):
    def __init__(self, *layers: nn.Module):
        super().__init__()
        self._submodules = nn.ModuleList(layers)

    def forward(self, x: torch.Tensor, h: Optional[dict] = None):
        if h is None:
            h = {}

        new_h = {}

        for i, layer in enumerate(self._submodules):
            layer_h = h.get(i, None)
            if layer_h is None:
                layer_result = layer(x)
            else:
                layer_result = layer(x, h[i])

            if isinstance(layer_result, tuple):
                if len(layer_result) == 2:
                    x, layer_new_h = layer_result
                else:
                    raise ValueError(
                        f"The layer number {i} returned a tuple of length {len(layer_result)}."
                        f" A tensor or a tuple of two elements was expected."
                    )
            elif isinstance(layer_result, torch.Tensor):
                x = layer_result
                layer_new_h = None
            else:
                raise TypeError(
                    f"The layer number {i} returned an object of type {type(layer_result)}."
                    f" A tensor or a tuple of two elements was expected."
                )

            if layer_new_h is not None:
                new_h[i] = layer_new_h

        if len(new_h) == 0:
            return x
        else:
            return x, new_h

    def __iter__(self):
        return self._submodules.__iter__()

    def __getitem__(self, i):
        return self._submodules[i]

    def append(self, module: nn.Module):
        self._submodules.append(module)



# Copyright 2022 NNAISENSE SA
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import itertools
import pickle
from collections import namedtuple
from copy import copy, deepcopy

import numpy as np
import pytest
import torch

from evotorch import Problem
from evotorch.distributions import ExpGaussian, ExpSeparableGaussian, SeparableGaussian, SymmetricSeparableGaussian
from evotorch.testing import assert_allclose, assert_eachclose
from evotorch.tools import ObjectArray, ReadOnlyTensor, make_tensor
from evotorch.tools.cloning import deep_clone
from evotorch.tools.immutable import ImmutableDict, ImmutableList


def test_deep_cloning():
    tolerance = 1e-4

    x = torch.randn(1000, 10)
    y = x[2:5]
    a = x[2:5]
    b = x[12:15]

    data = (y, {"a": a, "b": b})

    clone_of_data = deep_clone(data, otherwise_fail=True)
    clone_of_y = clone_of_data[0]
    clone_of_a = clone_of_data[1]["a"]
    clone_of_b = clone_of_data[1]["b"]

    assert_allclose(clone_of_y, y, atol=tolerance)
    assert_allclose(clone_of_a, a, atol=tolerance)
    assert_allclose(clone_of_b, b, atol=tolerance)

    # The deep-clone operation detaches the tensors from their original storages.
    # Therefore, unlike their originals y and a, clone_of_y and clone_of_a should NOT share storage.
    clone_of_y[:] = -1
    clone_of_a[:] = 1

    assert_eachclose(clone_of_y, -1, atol=tolerance)
    assert_eachclose(clone_of_a, 1, atol=tolerance)


def test_objectarray_cloning():
    tolerance = 1e-4

    pytorch_tensor = torch.randn(20)
    numpy_array = np.random.randn(5)

    original_data = [
        "a",
        {
            1: [10, 20],
            2: [20, 40],
            3: numpy_array,
        },
        pytorch_tensor,
    ]

    arr = make_tensor(original_data, dtype=object)

    # Test whether or not the data was properly cloned while putting into the ObjectArray.
    # The elements in the ObjectArray are expected to have the same values with their originals, while not being
    # the same objects.
    assert arr[0] == "a"

    assert isinstance(arr[1], ImmutableDict)
    assert arr[1][1] == [10, 20]
    assert arr[1][1] is not original_data[1][1]
    assert isinstance(arr[1][1], ImmutableList)

    assert arr[1][2] == [20, 40]
    assert arr[1][2] is not original_data[1][2]
    assert isinstance(arr[1][2], ImmutableList)

    assert_allclose(arr[1][3], numpy_array, atol=tolerance)
    assert arr[1][3] is not numpy_array
    assert not arr[1][3].flags["WRITEABLE"]  # The clone of numpy_array is expected to be read-only

    assert_allclose(arr[2], pytorch_tensor, atol=tolerance)
    assert arr[2] is not pytorch_tensor
    assert isinstance(arr[2], ReadOnlyTensor)  # The clone of pytorch_tensor is expected to be read-only

    # Now we use the `.clone()` method to obtain a mutable clone of the ObjectArray
    mutable_clone = arr.clone()

    # The elements in the newly made mutable clone are expected to be copies of the original data.
    # These elements are expected to have the same values with their originals, while not being the same objects.

    assert isinstance(mutable_clone, np.ndarray)
    assert mutable_clone.flags["WRITEABLE"]
    assert mutable_clone[0] == "a"

    assert isinstance(mutable_clone[1], dict)
    assert mutable_clone[1][1] == [10, 20]
    assert isinstance(mutable_clone[1][1], list)
    assert mutable_clone[1][1] is not arr[1][1]
    assert mutable_clone[1][1] is not original_data[1][1]

    assert mutable_clone[1][2] == [20, 40]
    assert isinstance(mutable_clone[1][2], list)
    assert mutable_clone[1][2] is not arr[1][2]
    assert mutable_clone[1][2] is not original_data[1][2]

    assert_allclose(mutable_clone[1][3], numpy_array, atol=tolerance)
    assert mutable_clone[1][3] is not arr[1][3]
    assert mutable_clone[1][3] is not numpy_array

    assert_allclose(mutable_clone[2], pytorch_tensor, atol=tolerance)
    assert mutable_clone[2] is not arr[2]
    assert mutable_clone[2] is not pytorch_tensor


def _assert_equals_but_not_same(a, b, atol=None):
    assert a is not b
    if atol is None:
        assert a == b
    else:
        assert_allclose(a, b, atol=atol)


class CloningMethods:
    @staticmethod
    def clone_via_method(x):
        return x.clone()

    @staticmethod
    def clone_via_copy(x):
        return copy(x)

    @staticmethod
    def clone_via_deepcopy(x):
        return deepcopy(x)

    @staticmethod
    def clone_via_clone_func(x):
        from evotorch.tools import clone

        return clone(x)

    @staticmethod
    def deep_clone(x):
        return deep_clone(x, otherwise_deepcopy=True)

    @staticmethod
    def pickle_and_unpickle(x):
        return pickle.loads(pickle.dumps(x))


@pytest.mark.parametrize(
    "clone_func",
    [
        CloningMethods.clone_via_clone_func,
        CloningMethods.deep_clone,
    ],
)
def test_cloning_container_with_cycle(clone_func):
    tolerance = 1e-4

    pytorch_tensor = torch.randn(5)

    a = {"a": 1, "b": pytorch_tensor, "c": [1, 2], "d": pytorch_tensor}
    a["c"].append(a)
    a["e"] = a["c"]

    MyNamedTuple = namedtuple("MyNamedTuple", ("aa", "bb"))
    a["f"] = MyNamedTuple(100, 200)
    a["g"] = a["f"]

    clone_of_a = clone_func(a)

    assert a is not clone_of_a

    assert set(a.keys()) == set(clone_of_a.keys())

    assert a["a"] == clone_of_a["a"]
    _assert_equals_but_not_same(a["b"], clone_of_a["b"], atol=tolerance)

    assert a["c"] is not clone_of_a["c"]
    assert a["c"][:2] == clone_of_a["c"][:2]
    assert clone_of_a["c"][2] is clone_of_a

    # Assert that the namedtuple was cloned correctly
    assert clone_of_a["f"] is not a["f"]
    assert clone_of_a["g"] is not a["g"]
    assert clone_of_a["f"] == a["f"]
    assert set(clone_of_a["f"]._fields) == set(a["f"]._fields)

    # Some elements in the original object are expected to be referred to twice, just like how their original
    # counterparts were referred to twice.
    assert clone_of_a["b"] is clone_of_a["d"]
    assert clone_of_a["c"] is clone_of_a["e"]
    assert clone_of_a["f"] is clone_of_a["g"]


@pytest.mark.parametrize(
    "clone_func",
    [
        CloningMethods.clone_via_clone_func,
        CloningMethods.clone_via_copy,
        CloningMethods.clone_via_deepcopy,
        CloningMethods.clone_via_method,
        CloningMethods.deep_clone,
        CloningMethods.pickle_and_unpickle,
    ],
)
def test_cloning_problem_with_cycle(clone_func):
    tolerance = 1e-4

    problem = Problem(
        "min",
        torch.linalg.norm,
        initial_bounds=(-1.0, 1.0),
        solution_length=5,
        dtype="float32",
    )

    some_tensor = torch.FloatTensor([10, 20, 30])
    some_data = (some_tensor[:], [1, 2, problem])
    problem.some_tensor = some_tensor
    problem.some_data = some_data

    clone_of_problem = clone_func(problem)

    assert_allclose(clone_of_problem.some_tensor, problem.some_tensor, atol=tolerance)
    assert_allclose(clone_of_problem.some_data[0], problem.some_data[0], atol=tolerance)
    assert_allclose(clone_of_problem.some_tensor, clone_of_problem.some_data[0], atol=tolerance)

    # Check whether or not the cloning was done correctly
    assert len(clone_of_problem.some_data) == len(problem.some_data)
    assert len(clone_of_problem.some_data[-1]) == len(problem.some_data[-1])
    assert clone_of_problem.some_data[-1][0] == problem.some_data[-1][0]
    assert clone_of_problem.some_data[-1][1] == problem.some_data[-1][1]

    # Although their values are the same, the inner list of `some_data` must be a clone (not the same object)
    assert clone_of_problem.some_data[-1] is not problem.some_data[-1]

    # The clone of the problem object is expected to point to itself like the original one does
    assert clone_of_problem is not problem
    assert clone_of_problem.some_data[-1][-1] is clone_of_problem

    # After the cloning, we expect that the tensors are independent: they do not share storage anymore
    clone_of_problem.some_tensor[:] = -1
    clone_of_problem.some_data[0][:] = 1
    assert_eachclose(clone_of_problem.some_tensor, -1, atol=tolerance)
    assert_eachclose(clone_of_problem.some_data[0], 1, atol=tolerance)


@pytest.mark.parametrize(
    "clone_func,expect_mutable",
    [
        (CloningMethods.clone_via_clone_func, False),
        (CloningMethods.clone_via_copy, False),
        (CloningMethods.clone_via_deepcopy, False),
        (CloningMethods.clone_via_method, True),
        (CloningMethods.deep_clone, False),
        (CloningMethods.pickle_and_unpickle, False),
    ],
)
def test_cloning_objectarray_with_cycles(clone_func, expect_mutable):
    # In this test, we prepare a list with self cycles in it.
    # We first clone it by putting into an ObjectArray.
    # We then clone this ObjectArray again by using various cloning methods.
    # If all of our cloning methods are implemented correctly, we will get the expected results listed below.

    tolerance = 1e-4

    pytorch_tensor = torch.randn(20)
    numpy_array = np.random.randn(5)

    original_data = [
        "one",
        "two",
        pytorch_tensor,
    ]

    inside_dict = {
        "a": 1,
        "b": 2,
        "c": numpy_array,
    }

    inside_dict["d"] = inside_dict

    original_data.append(inside_dict)
    original_data.append(original_data)
    original_data.append(numpy_array)

    # The final state of original_data is:
    #
    # [
    #     "one",                         # 0
    #     "two",                         # 1
    #     pytorch_tensor,                # 2
    #     {                              # 3
    #         "a": 1,
    #         "b": 2,
    #         "c": numpy_array,
    #         "d": <the dict itself>,
    #     },
    #     <the list itself>,             # 4
    #     numpy_array,                   # 5
    # ]

    arr = make_tensor(original_data, dtype=object)
    clone_of_arr = clone_func(arr)

    # The clone of the array is expected to have the same number of elements
    assert len(clone_of_arr) == len(arr)

    # The clone of the inner dictionary is expected to have the same keys
    assert set(clone_of_arr[3].keys()) == set(arr[3].keys())

    # The integers stored via the keys "a" and "b" in the cloned dictionary must be equal
    assert clone_of_arr[3]["a"] == arr[3]["a"]
    assert clone_of_arr[3]["b"] == arr[3]["b"]

    # The numpy array was referred to in the original array twice.
    # We expect that the clone of the original array also refers to its own numpy array clone twice.
    assert isinstance(clone_of_arr[3]["c"], np.ndarray)
    assert isinstance(clone_of_arr[5], np.ndarray)
    assert clone_of_arr[3]["c"] is clone_of_arr[5]

    # We expect that the numpy array was also cloned. It is not the same with its original.
    assert_allclose(clone_of_arr[5], numpy_array, atol=tolerance)
    assert clone_of_arr[5] is not numpy_array
    assert clone_of_arr[5] is not arr[5]

    # We expect that the pytorch tensor was also cloned. It is not the same with its original.
    assert_allclose(clone_of_arr[2], pytorch_tensor, atol=tolerance)
    assert isinstance(clone_of_arr[2], torch.Tensor)
    assert clone_of_arr[2] is not pytorch_tensor
    assert clone_of_arr[2] is not arr[2]

    if expect_mutable:
        assert isinstance(clone_of_arr, np.ndarray)
        assert not isinstance(clone_of_arr[2], ReadOnlyTensor)
        assert isinstance(clone_of_arr[3], dict)
    else:
        assert isinstance(clone_of_arr, ObjectArray)
        assert isinstance(clone_of_arr[2], ReadOnlyTensor)
        assert isinstance(clone_of_arr[3], ImmutableDict)
        assert isinstance(clone_of_arr[4], ObjectArray)
        assert isinstance(clone_of_arr[5], np.ndarray)

        if clone_func is not CloningMethods.pickle_and_unpickle:
            assert not clone_of_arr[5].flags["WRITEABLE"]


@pytest.mark.parametrize(
    "distcls,clone_func",
    itertools.product(
        (SeparableGaussian, SymmetricSeparableGaussian, ExpGaussian, ExpSeparableGaussian),
        (
            CloningMethods.clone_via_clone_func,
            CloningMethods.clone_via_copy,
            CloningMethods.clone_via_deepcopy,
            CloningMethods.clone_via_method,
            CloningMethods.pickle_and_unpickle,
        ),
    ),
)
def test_cloning_gaussian_distribution(distcls, clone_func):
    tolerance = 1e-4

    n = 20
    mu = torch.randn(n)
    sigma = torch.randn(n)
    params = {"mu": mu, "sigma": sigma}

    dist = distcls(params)
    clone_of_dist = clone_func(dist)

    mu = dist.mu
    sigma = dist.sigma

    clone_of_mu = clone_of_dist.mu
    clone_of_sigma = clone_of_dist.sigma

    _assert_equals_but_not_same(clone_of_mu, mu, atol=tolerance)
    _assert_equals_but_not_same(clone_of_sigma, sigma, atol=tolerance)



# Copyright 2022 NNAISENSE SA
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Module defining decorators for evotorch."""

from typing import Callable


def pass_info(fn_or_class: Callable) -> Callable:
    """
    Decorates a callable so that the neuroevolution problem class (e.g. GymNE) will
    pass information regarding the task at hand, in the form of keyword arguments.

    For example, in the case of [GymNE][evotorch.neuroevolution.GymNE] or
    [VecGymNE][evotorch.neuroevolution.VecGymNE], the passed information would
    include dimensions of the observation and action spaces.

    Example:
        ```python
        @pass_info
        class MyModule(nn.Module):
            def __init__(self, obs_length: int, act_length: int, **kwargs):
                # Because MyModule is decorated with @pass_info, it receives
                # keyword arguments related to the environment "CartPole-v0",
                # including obs_length and act_length.
                ...


        problem = GymNE(
            "CartPole-v0",
            network=MyModule,
            ...,
        )
        ```

    Args:
        fn_or_class (Callable): Function or class to decorate

    Returns:
        Callable: Decorated function or class
    """
    setattr(fn_or_class, "__evotorch_pass_info__", True)
    return fn_or_class



# Copyright 2022 NNAISENSE SA
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
This namespace provides various vectorized reinforcement learning utilities.
"""


from collections.abc import Mapping
from copy import deepcopy
from typing import Any, Callable, Iterable, Optional, Union

import gym
import numpy as np
import torch
from functorch import vmap
from gym.spaces import Box, Discrete, Space
from gym.vector import SyncVectorEnv
from torch import nn
from torch.nn import utils as nnu

try:
    import jax
    import jax.dlpack
    from jax import numpy as jnp
except ImportError:
    jax = None
    jnp = None


if jax is not None:
    JaxArray = jnp.ndarray

    def is_jax_array(x: Any) -> bool:
        return isinstance(x, JaxArray)

    def jax_to_torch(x: JaxArray) -> torch.Tensor:
        x = jax.dlpack.to_dlpack(x)
        x = torch.utils.dlpack.from_dlpack(x)
        return x

    def torch_to_jax(x: torch.Tensor) -> JaxArray:
        x = torch.utils.dlpack.to_dlpack(x)
        x = jax.dlpack.from_dlpack(x)
        return x

else:

    def _jax_is_missing():
        raise ImportError("The module `jax` is missing.")

    class JaxArray:
        def __init__(self, *args, **kwargs):
            _jax_is_missing()

    def is_jax_array(x: Any) -> bool:
        return False

    def jax_to_torch(x: JaxArray) -> torch.Tensor:
        _jax_is_missing()

    def torch_to_jax(x: torch.Tensor) -> JaxArray:
        _jax_is_missing()


try:
    import brax
    import brax.envs
except ImportError:
    brax = None


if brax is not None:
    from brax.envs import Env as BraxEnv

    def is_brax_env(env: Any) -> bool:
        return isinstance(env, BraxEnv)

else:

    def _brax_is_missing():
        raise ImportError("The module `brax` is missing.")

    class BraxEnv:
        def __init__(self, *args, **kwargs):
            _brax_is_missing()

    def is_brax_env(env: Any) -> bool:
        return False


def array_type(x: Any, fallback: Optional[str] = None) -> str:
    """
    Get the type of an array as a string ("jax", "torch", or "numpy").
    If the type of the array cannot be determined and a fallback is provided,
    then the fallback value will be returned.

    Args:
        x: The array whose type will be determined.
        fallback: Fallback value, as a string, which will be returned if the
            array type cannot be determined.
    Returns:
        The array type as a string ("jax", "torch", or "numpy").
    Raises:
        TypeError: if the array type cannot be determined and a fallback
            value is not provided.
    """
    if is_jax_array(x):
        return "jax"
    elif isinstance(x, torch.Tensor):
        return "torch"
    elif isinstance(x, np.ndarray):
        return "numpy"
    elif fallback is not None:
        return fallback
    else:
        raise TypeError(f"The object has an unrecognized type: {type(x)}")


def convert_from_torch(x: torch.Tensor, array_type: str) -> Any:
    """
    Convert the given PyTorch tensor to an array of the specified type.

    Args:
        x: The PyTorch array that will be converted.
        array_type: Type to which the PyTorch tensor will be converted.
            Expected as one of these strings: "jax", "torch", "numpy".
    Returns:

    """
    if array_type == "torch":
        return x
    elif array_type == "jax":
        return torch_to_jax(x)
    elif array_type == "numpy":
        return x.cpu().numpy()
    else:
        raise ValueError(f"Unrecognized array type: {array_type}")


def convert_to_torch(x: Any) -> torch.Tensor:
    """
    Convert the given array to PyTorch tensor.

    Args:
        x: Array to be converted. Can be a JAX array, a numpy array,
            a PyTorch tensor (in which case the input tensor will be
            returned as it is) or any Iterable object.
    Returns:
        The PyTorch counterpart of the given array.
    """
    if isinstance(x, torch.Tensor):
        return x
    elif is_jax_array(x):
        return jax_to_torch(x)
    elif isinstance(x, np.ndarray):
        return torch.from_numpy(x)
    else:
        return torch.as_tensor(x)


def _must_be_supported_space(space: Space):
    if not isinstance(space, (Box, Discrete)):
        raise TypeError(
            f"Actions and observations must be in one of these spaces: Box, Discrete."
            f" Encountered an unsupported space: {type(space)}."
        )


def convert_to_torch_bool(x: Any) -> torch.Tensor:
    """
    Convert the given array to a PyTorch tensor of bools.

    If the given object is an array of floating point numbers, then, values
    that are near to 0.0 (with a tolerance of 1e-4) will be converted to
    False, and the others will be converted to True.
    If the given object is an array of integers, then zero values will be
    converted to False, and non-zero values will be converted to True.
    If the given object is an array of booleans, then no change will be made
    to those boolean values.

    The given object can be a JAX array, a numpy array, or a PyTorch tensor.
    The result will always be a PyTorch tensor.

    Args:
        x: Array to be converted.
    Returns:
        The array converted to a PyTorch tensor with its dtype set as bool.
    """
    x = convert_to_torch(x)
    if x.dtype == torch.bool:
        pass  # nothing to do
    elif "float" in str(x.dtype):
        x = torch.abs(x) > 1e-4
    else:
        x = torch.as_tensor(x, dtype=torch.bool)

    return x


class TorchWrapper(gym.Wrapper):
    """
    A gym wrapper which ensures that the actions, observations, rewards, and
    the 'done' values are expressed as PyTorch tensors.
    """

    def __init__(
        self,
        env: Union[gym.Env],
        *,
        force_classic_api: bool = False,
        discrete_to_continuous_act: bool = False,
        clip_actions: bool = False,
        **kwargs,
    ):
        """
        `__init__(...)`: Initialize the TorchWrapper.

        Args:
            env: The gym environment to be wrapped.
            force_classic_api: Set this as True if you would like to enable
                the classic API. In the classic API, the `reset(...)` method
                returns only the observation and the `step(...)` method
                returns 4 elements (not 5).
            discrete_to_continuous_act: When this is set as True and the
                wrapped environment has a Discrete action space, this wrapper
                will transform the action space to Box. A Discrete-action
                environment with `n` actions will be converted to a Box-action
                environment where the action length is `n`.
                The index of the largest value within the action vector will
                be applied to the underlying environment.
            clip_actions: Set this as True if you would like to clip the given
                actions so that they conform to the declared boundaries of the
                action space.
            kwargs: Expected in the form of additional keyword arguments.
                These additional keyword arguments are passed to the
                superclass.
        """
        super().__init__(env, **kwargs)

        # Declare the variable that will store the array type of the underlying environment.
        self.__array_type: Optional[str] = None

        if hasattr(env, "single_observation_space"):
            # If the underlying environment has the attribute "single_observation_space",
            # then this is a vectorized environment.
            self.__vectorized = True

            # Get the observation and action spaces.
            obs_space = env.single_observation_space
            act_space = env.single_action_space
        else:
            # If the underlying environment has the attribute "single_observation_space",
            # then this is a non-vectorized environment.
            self.__vectorized = False

            # Get the observation and action spaces.
            obs_space = env.observation_space
            act_space = env.action_space

        # Ensure that the observation and action spaces are supported.
        _must_be_supported_space(obs_space)
        _must_be_supported_space(act_space)

        # Store the choice of the user regarding "force_classic_api".
        self.__force_classic_api = bool(force_classic_api)

        if isinstance(act_space, Discrete) and discrete_to_continuous_act:
            # The underlying action space is Discrete and `discrete_to_continuous_act` is given as True.
            # Therefore, we convert the action space to continuous (to Box).

            # Take the shape and the dtype of the discrete action space.
            single_action_shape = (act_space.n,)
            single_action_dtype = torch.from_numpy(np.array([], dtype=act_space.dtype)).dtype

            # We store the integer dtype of the environment.
            self.__discrete_dtype = single_action_dtype

            if self.__vectorized:
                # If the environment is vectorized, we declare the new `action_space` and the `single_action_space`
                # for the enviornment.
                action_shape = (env.num_envs,) + single_action_shape
                self.single_action_space = Box(float("-inf"), float("inf"), shape=single_action_shape, dtype=np.float32)
                self.action_space = Box(float("-inf"), float("inf"), shape=action_shape, dtype=np.float32)
            else:
                # If the environment is not vectorized, we declare the new `action_space` for the environment.
                self.action_space = Box(float("-inf"), float("inf"), shape=single_action_shape, dtype=np.float32)
        else:
            # This is the case where we do not transform the action space.
            # The discrete dtype will not be used, so, we set it as None.
            self.__discrete_dtype = None

        if isinstance(act_space, Box) and clip_actions:
            # If the action space is Box and the wrapper is configured to clip the actions, then we store the lower
            # and the upper bounds for the actions.
            self.__act_lb = torch.from_numpy(act_space.low)
            self.__act_ub = torch.from_numpy(act_space.high)
        else:
            # If there will not be any action clipping, then we store the lower and the upper bounds as None.
            self.__act_lb = None
            self.__act_ub = None

    @property
    def array_type(self) -> Optional[str]:
        """
        Get the array type of the wrapped environment.
        This can be "jax", "torch", or "numpy".
        """
        return self.__array_type

    def __infer_array_type(self, observation):
        if self.__array_type is None:
            # If the array type is not determined yet, set it as the array type of the received observation.
            # If the observation has an unrecognized type, set the array type as "numpy".
            self.__array_type = array_type(observation, "numpy")

    def reset(self, *args, **kwargs):
        """Reset the environment"""

        # Call the reset method of the wrapped environment.
        reset_result = self.env.reset(*args, **kwargs)

        if isinstance(reset_result, tuple):
            # If we received a tuple of two elements, then we assume that this is the new gym API.
            # We note that we received an info dictionary.
            got_info = True
            # We keep the received observation and info.
            observation, info = reset_result
        else:
            # If we did not receive a tuple, then we assume that this is the old gym API.
            # We note that we did not receive an info dictionary.
            got_info = False
            # We keep the received observation.
            observation = reset_result
            # We did not receive an info dictionary, so, we set it as an empty dictionary.
            info = {}

        # We understand the array type of the underlying environment from the first observation.
        self.__infer_array_type(observation)

        # Convert the observation to a PyTorch tensor.
        observation = convert_to_torch(observation)

        if self.__force_classic_api:
            # If the option `force_classic_api` was set as True, then we only return the observation.
            return observation
        else:
            # Here we handle the case where `force_classic_api` was set as False.
            if got_info:
                # If we got an additional info dictionary, we return it next to the observation.
                return observation, info
            else:
                # If we did not get any info dictionary, we return only the observation.
                return observation

    def step(self, action, *args, **kwargs):
        """Take a step in the environment"""

        if self.__array_type is None:
            # If the array type is not known yet, then probably `reset()` has not been called yet.
            # We raise an error.
            raise ValueError(
                "Could not understand what type of array this environment works with."
                " Perhaps the `reset()` method has not been called yet?"
            )

        if self.__discrete_dtype is not None:
            # If the wrapped environment is discrete-actioned, then we take the integer counterpart of the action.
            action = torch.argmax(action, dim=-1).to(dtype=self.__discrete_dtype)

        if self.__act_lb is not None:
            # The internal variable `__act_lb` having a value other than None means that the initialization argument
            # `clip_actions` was given as True.
            # Therefore, we clip the actions.
            self.__act_lb = self.__act_lb.to(action.device)
            self.__act_ub = self.__act_ub.to(action.device)
            action = torch.max(action, self.__act_lb)
            action = torch.min(action, self.__act_ub)

        # Convert the action tensor to the expected array type of the underlying environment.
        action = convert_from_torch(action, self.__array_type)

        # Perform the step and get the result.
        result = self.env.step(action, *args, **kwargs)

        if not isinstance(result, tuple):
            # If the `step(...)` method returned anything other than tuple, we raise an error.
            raise TypeError(f"Expected a tuple as the result of the `step()` method, but received a {type(result)}")

        if len(result) == 5:
            # If the result is a tuple of 5 elements, then we note that we are using the new API.
            using_new_api = True
            # Take the observation, reward, two boolean variables done and done2 indicating that the episode(s)
            # has/have ended, and additional info.
            # `done` indicates whether or not the episode(s) reached terminal state(s).
            # `done2` indicates whether or not the episode(s) got truncated because of the timestep limit.
            observation, reward, done, done2, info = result
        elif len(result) == 4:
            # If the result is a tuple of 5 elements, then we note that we are not using the new API.
            using_new_api = False
            # Take the observation, reward, the done boolean flag, and additional info.
            observation, reward, done, info = result
            done2 = None
        else:
            raise ValueError(f"Unexpected number of elements were returned from step(): {len(result)}")

        # Convert the observation, reward, and done variables to PyTorch tensors.
        observation = convert_to_torch(observation)
        reward = convert_to_torch(reward)
        done = convert_to_torch_bool(done)
        if done2 is not None:
            done2 = convert_to_torch_bool(done2)

        if self.__force_classic_api:
            # This is the case where the initialization argument `force_classic_api` was set as True.
            if done2 is not None:
                # We combine the terminal state and truncation signals into a single boolean tensor indicating
                # whether or not the episode(s) ended.
                done = done | done2
            # Return 4 elements, compatible with the classic gym API.
            return observation, reward, done, info
        else:
            # This is the case where the initialization argument `force_classic_api` was set as False.
            if using_new_api:
                # If we are using the new API, then we return the 5-element result.
                return observation, reward, done, done2, info
            else:
                # If we are using the new API, then we return the 4-element result.
                return observation, reward, done, info


def make_brax_env(
    env_name: str,
    *,
    force_classic_api: bool = False,
    num_envs: Optional[int] = None,
    discrete_to_continuous_act: bool = False,
    clip_actions: bool = False,
    **kwargs,
) -> TorchWrapper:
    """
    Make a brax environment and wrap it via TorchWrapper.

    Args:
        env_name: Name of the brax environment, as string (e.g. "humanoid").
        force_classic_api: Whether or not the classic gym API is to be used.
        num_envs: Batch size for the vectorized environment.
        discrete_to_continuous_act: Whether or not the the discrete action
            space of the environment is to be converted to a continuous one.
            This does nothing if the environment's action space is not
            discrete.
        clip_actions: Whether or not the actions should be explicitly clipped
            so that they stay within the declared action boundaries.
        kwargs: Expected in the form of additional keyword arguments, these
            are passed to the environment.
    Returns:
        The brax environment, wrapped by TorchWrapper.
    """

    if brax is not None:
        config = {}
        config.update(kwargs)
        if num_envs is not None:
            config["batch_size"] = num_envs
        env = brax.envs.create_gym_env(env_name, **config)
        env = TorchWrapper(
            env,
            force_classic_api=force_classic_api,
            discrete_to_continuous_act=discrete_to_continuous_act,
            clip_actions=clip_actions,
        )
        return env
    else:
        _brax_is_missing()


def make_gym_env(
    env_name: str,
    *,
    force_classic_api: bool = False,
    num_envs: Optional[int] = None,
    discrete_to_continuous_act: bool = False,
    clip_actions: bool = False,
    **kwargs,
) -> TorchWrapper:
    """
    Make gym environments and wrap them via a SyncVectorEnv and a TorchWrapper.

    Args:
        env_name: Name of the gym environment, as string (e.g. "Humanoid-v4").
        force_classic_api: Whether or not the classic gym API is to be used.
        num_envs: Batch size for the vectorized environment.
        discrete_to_continuous_act: Whether or not the the discrete action
            space of the environment is to be converted to a continuous one.
            This does nothing if the environment's action space is not
            discrete.
        clip_actions: Whether or not the actions should be explicitly clipped
            so that they stay within the declared action boundaries.
        kwargs: Expected in the form of additional keyword arguments, these
            are passed to the environment.
    Returns:
        The gym environments, wrapped by a TorchWrapper.
    """

    def make_the_env():
        return gym.make(env_name, **kwargs)

    env_fns = [make_the_env for _ in range(num_envs)]
    vec_env = TorchWrapper(
        SyncVectorEnv(env_fns),
        force_classic_api=force_classic_api,
        discrete_to_continuous_act=discrete_to_continuous_act,
        clip_actions=clip_actions,
    )

    return vec_env


def make_vector_env(
    env_name: str,
    *,
    force_classic_api: bool = False,
    num_envs: Optional[int] = None,
    discrete_to_continuous_act: bool = False,
    clip_actions: bool = False,
    **kwargs,
) -> TorchWrapper:
    """
    Make a new vectorized environment and wrap it via TorchWrapper.

    Args:
        env_name: Name of the gym environment, as string.
            If the string starts with "gym::" (e.g. "gym::Humanoid-v4", etc.),
            then it is assumed that the target environment is a classical gym
            environment which will first be wrapped via a SyncVectorEnv and
            then via a TorchWrapper.
            If the string starts with "brax::" (e.g. "brax::humanoid", etc.),
            then it is assumed that the target environment is a brax
            environment which will be wrapped via TorchWrapper.
            If the string does not contain "::" at all (e.g. "Humanoid-v4"),
            then it is assumed that the target environment is a classical gym
            environment. Therefore, "gym::Humanoid-v4" and "Humanoid-v4"
            are equivalent.
        force_classic_api: Whether or not the classic gym API is to be used.
        num_envs: Batch size for the vectorized environment.
        discrete_to_continuous_act: Whether or not the the discrete action
            space of the environment is to be converted to a continuous one.
            This does nothing if the environment's action space is not
            discrete.
        clip_actions: Whether or not the actions should be explicitly clipped
            so that they stay within the declared action boundaries.
        kwargs: Expected in the form of additional keyword arguments, these
            are passed to the environment.
    Returns:
        The gym environments, wrapped by a TorchWrapper.
    """

    env_parts = str(env_name).split("::", maxsplit=1)

    if len(env_parts) == 0:
        raise ValueError(f"Invalid value for `env_name`: {repr(env_name)}")
    elif len(env_parts) == 1:
        fn = make_gym_env
    elif len(env_parts) == 2:
        env_name = env_parts[1]
        if env_parts[0] == "gym":
            fn = make_gym_env
        elif env_parts[0] == "brax":
            fn = make_brax_env
        else:
            invalid_value = env_parts[0] + "::"
            raise ValueError(
                f"The argument `env_name` starts with {repr(invalid_value)}, implying that the environment is stored"
                f" in a registry named {repr(env_parts[0])}."
                f" However, the registry {repr(env_parts[0])} is not recognized."
                f" Supported environment registries are: 'gym', 'brax'."
            )
    else:
        assert False, "Unexpected value received from len(env_parts)"

    return fn(
        env_name,
        force_classic_api=force_classic_api,
        num_envs=num_envs,
        discrete_to_continuous_act=discrete_to_continuous_act,
        clip_actions=clip_actions,
        **kwargs,
    )


MaskOrIndices = Union[int, Iterable]


def reset_tensors(x: Any, indices: MaskOrIndices):
    """
    Reset the specified regions of the given tensor(s) as 0.

    Note that the resetting is performed in-place, which means, the provided tensors are modified.

    The regions are determined by the argument `indices`, which can be a sequence of booleans (in which case it is
    interpreted as a mask), or a sequence of integers (in which case it is interpreted as the list of indices).

    For example, let us imagine that we have the following tensor:

    ```python
    import torch

    x = torch.tensor(
        [
            [0, 1, 2, 3],
            [4, 5, 6, 7],
            [8, 9, 10, 11],
            [12, 13, 14, 15],
        ],
        dtype=torch.float32,
    )
    ```

    If we wish to reset the rows with indices 0 and 2, we could use:

    ```python
    reset_tensors(x, [0, 2])
    ```

    The new value of `x` would then be:

    ```
    torch.tensor(
        [
            [0, 0, 0, 0],
            [4, 5, 6, 7],
            [0, 0, 0, 0],
            [12, 13, 14, 15],
        ],
        dtype=torch.float32,
    )
    ```

    The first argument does not have to be a single tensor.
    Instead, it can be a container (i.e. a dictionary-like object or an iterable) that stores tensors.
    In this case, each tensor stored by the container will be subject to resetting.
    In more details, each tensor within the iterable(s) and each tensor within the value part of the dictionary-like
    object(s) will be reset.

    As an example, let us assume that we have the following collection:

    ```python
    a = torch.tensor(
        [
            [0, 1],
            [2, 3],
            [4, 5],
        ],
        dtype=torch.float32,
    )

    b = torch.tensor(
        [
            [0, 10, 20],
            [30, 40, 50],
            [60, 70, 80],
        ],
        dtype=torch.float32,
    )

    c = torch.tensor(
        [
            [100],
            [200],
            [300],
        ],
        dtype=torch.float32,
    )

    d = torch.tensor([-1, -2, -3], dtype=torch.float32)

    my_tensors = [a, {"1": b, "2": (c, d)}]
    ```

    To clear the regions with indices, e.g, (1, 2), we could do:

    ```python
    reset_tensors(my_tensors, [1, 2])
    ```

    and the result would be:

    ```
    >>> print(a)
    torch.tensor(
        [
            [0, 1],
            [0, 0],
            [0, 0],
        ],
        dtype=torch.float32,
    )

    >>> print(b)
    torch.tensor(
        [
            [0, 10, 20],
            [0, 0, 0],
            [0, 0, 0],
        ],
        dtype=torch.float32,
    )

    >>> print(c)
    c = torch.tensor(
        [
            [100],
            [0],
            [0],
        ],
        dtype=torch.float32,
    )

    >>> print(d)
    torch.tensor([-1, 0, 0], dtype=torch.float32)
    ```

    Args:
        x: A tensor or a collection of tensors, whose values are subject to resetting.
        indices: A sequence of integers or booleans, specifying which regions of the tensor(s) will be reset.
    """
    if isinstance(x, torch.Tensor):
        # If the first argument is a tensor, then we clear it according to the indices we received.
        x[indices] = 0
    elif isinstance(x, (str, bytes, bytearray)):
        # str, bytes, and bytearray are the types of `Iterable` that we do not wish to process.
        # Therefore, we explicitly add a condition for them here, and explicitly state that nothing should be done
        # when instances of them are encountered.
        pass
    elif isinstance(x, Mapping):
        # If the first argument is a Mapping (i.e. a dictionary-like object), then, for each value part of the
        # Mapping instance, we call this function itself.
        for key, value in x.items():
            reset_tensors(value, indices)
    elif isinstance(x, Iterable):
        # If the first argument is an Iterable (e.g. a list, a tuple, etc.), then, for each value contained by this
        # Iterable instance, we call this function itself.
        for value in x:
            reset_tensors(value, indices)


class Policy:
    """
    A Policy for deciding the actions for a reinforcement learning environment.

    This can be seen as a stateful wrapper around a PyTorch module.

    Let us assume that we have the following PyTorch module:

    ```python
    from torch import nn

    net = nn.Linear(5, 8)
    ```

    which has 48 parameters (when all the parameters are flattened).
    Let us randomly generate a parameter vector for our module `net`:

    ```python
    parameters = torch.randn(48)
    ```

    We can now prepare a policy:

    ```python
    policy = Policy(net)
    policy.set_parameters(parameters)
    ```

    If we generate a random observation:

    ```python
    observation = torch.randn(5)
    ```

    We can receive our action as follows:

    ```python
    action = policy(observation)
    ```

    If the PyTorch module that we wish to wrap is a recurrent network (i.e.
    a network which expects an optional second argument for the hidden state,
    and returns a second value which represents the updated hidden state),
    then, the hidden state is automatically managed by the Policy instance.

    Let us assume that we have a recurrent network named `recnet`.

    ```python
    policy = Policy(recnet)
    policy.set_parameters(parameters_of_recnet)
    ```

    In this case, because the hidden state of the network is internally
    managed, the usage is still the same with our previous non-recurrent
    example:

    ```python
    action = policy(observation)
    ```

    When using a recurrent module on multiple episodes, it is important
    to reset the hidden state of the network. This is achieved by the
    reset method:

    ```python
    policy.reset()
    action1 = policy(observation1)

    # action2 will be computed with the hidden state generated by the
    # previous forward-pass.
    action2 = policy(observation2)

    policy.reset()

    # action3 will be computed according to the renewed hidden state.
    action3 = policy(observation3)
    ```

    Both for non-recurrent and recurrent networks, it is possible to
    perform vectorized operations. For now, let us return to our
    first non-recurrent example:

    ```python
    net = nn.Linear(5, 8)
    ```

    Instead of generating only one parameter vector, we now generate
    a batch of parameter vectors. Let us say that our batch size is 10:

    ```python
    batch_of_parameters = torch.randn(10, 48)
    ```

    Like we did in the non-batched examples, we can do:

    ```python
    policy = Policy(net)
    policy.set_parameters(batch_of_parameters)
    ```

    Because we are now in the batched mode, `policy` now expects a batch
    of observations and will return a batch of actions:

    ```python
    batch_of_observations = torch.randn(10, 5)
    batch_of_actions = policy(batch_of_observations)
    ```

    When doing vectorized reinforcement learning with a recurrent module,
    it can be the case that only some of the environments are finished,
    and therefore it is necessary to reset the hidden states associated
    with those environments only. The `reset(...)` method of Policy
    has a second argument to specify which of the recurrent network
    instances are to be reset. For example, if the episodes of the
    environments with indices 2 and 5 are about to restart (and therefore
    we wish to reset the states of the networks with indices 2 and 5),
    then, we can do:

    ```python
    policy.reset(torch.tensor([2, 5]))
    ```
    """

    def __init__(self, net: Union[str, Callable, nn.Module], **kwargs):
        """
        `__init__(...)`: Initialize the Policy.

        Args:
            net: The network to be wrapped by the Policy object.
                This can be a string, a Callable (e.g. a `torch.nn.Module`
                subclass), or a `torch.nn.Module` instance.
                When this argument is a string, the network will be
                created with the help of the function
                `evotorch.neuroevolution.net.str_to_net(...)` and then
                wrapped. Please see the `str_to_net(...)` function's
                documentation for details regarding how a network structure
                can be expressed via strings.
            kwargs: Expected in the form of additional keyword arguments,
                these keyword arguments will be passed to the provided
                Callable object (if the argument `net` is a Callable)
                or to `str_to_net(...)` (if the argument `net` is a string)
                at the moment of generating the network.
                If the argument `net` is a `torch.nn.Module` instance,
                having any additional keyword arguments will trigger an
                error, because the network is already instantiated and
                therefore, it is not possible to pass these keyword arguments.
        """
        from ..net import str_to_net
        from ..net.functional import ModuleExpectingFlatParameters, make_functional_module

        if isinstance(net, str):
            self.__module = str_to_net(net, **kwargs)
        elif isinstance(net, nn.Module):
            if len(kwargs) > 0:
                raise ValueError(
                    f"When the network is given as an `nn.Module` instance, extra network arguments cannot be used"
                    f" (because the network is already instantiated)."
                    f" However, these extra keyword arguments were received: {kwargs}."
                )
            self.__module = net
        elif isinstance(net, Callable):
            self.__module = net(**kwargs)
        else:
            raise TypeError(
                f"The class `Policy` expected a string or an `nn.Module` instance, or a Callable, but received {net}"
                f" (whose type is {type(net)})."
            )

        self.__fmodule: ModuleExpectingFlatParameters = make_functional_module(self.__module)
        self.__state: Any = None
        self.__parameters: Optional[torch.Tensor] = None

    def set_parameters(self, parameters: torch.Tensor, indices: Optional[MaskOrIndices] = None, *, reset: bool = True):
        """
        Set the parameters of the policy.

        Args:
            parameters: A 1-dimensional or a 2-dimensional tensor containing
                the flattened parameters to be used with the neural network.
                If the given parameters are two-dimensional, then, given that
                the leftmost size of the parameter tensor is `n`, the
                observations will be expected in a batch with leftmost size
                `n`, and the returned actions will also be in a batch,
                again with the leftmost size `n`.
            indices: For when the parameters were previously given via a
                2-dimensional tensor, provide this argument if you would like
                to change only some rows of the previously given parameters.
                For example, if `indices` is given as `torch.tensor([2, 4])`
                and the argument `parameters` is given as a 2-dimensional
                tensor with leftmost size 2, then the rows with indices
                2 and 4 will be replaced by these new parameters provided
                via the argument `parameters`.
            reset: If given as True, the hidden states of the networks whose
                parameters just changed will be reset. If `indices` was not
                provided at all, then this means that the parameters of all
                networks are modified, in which case, all the hidden states
                will be reset.
                If given as False, no such resetting will be done.
        """
        if self.__parameters is None:
            if indices is not None:
                raise ValueError(
                    "The argument `indices` can be used only if network parameters were previously specified."
                    " However, it seems that the method `set_parameters(...)` was not called before."
                )
            self.__parameters = parameters
        else:
            if indices is None:
                self.__parameters = parameters
            else:
                self.__parameters[indices] = parameters

        if reset:
            self.reset(indices)

    def __call__(self, x: torch.Tensor) -> torch.Tensor:
        """
        Pass the given observations through the network.

        Args:
            x: The observations, as a PyTorch tensor.
                If the parameters were given (via the method
                `set_parameters(...)`) as a 1-dimensional tensor, then this
                argument is expected to store a single observation.
                If the parameters were given as a 2-dimensional tensor,
                then, this argument is expected to store a batch of
                observations, and the leftmost size of this observation
                tensor must match with the leftmost size of the parameter
                tensor.
        Returns:
            The output tensor, which represents the action to take.
        """
        if self.__parameters is None:
            raise ValueError("Please use the method `set_parameters(...)` before calling the policy.")

        if self.__state is None:
            further_args = (x,)
        else:
            further_args = (x, self.__state)

        parameters = self.__parameters
        ndim = parameters.ndim
        if ndim == 1:
            result = self.__fmodule(parameters, *further_args)
        elif ndim == 2:
            vmapped = vmap(self.__fmodule)
            result = vmapped(parameters, *further_args)
        else:
            raise ValueError(
                f"Expected the parameters as a 1 or 2 dimensional tensor."
                f" However, the received parameters tensor has {ndim} dimensions."
            )

        if isinstance(result, torch.Tensor):
            return result
        elif isinstance(result, tuple):
            result, state = result
            self.__state = state
            return result
        else:
            raise TypeError(f"The torch module used by the Policy returned an unexpected object: {result}")

    def reset(self, indices: Optional[MaskOrIndices] = None, *, copy: bool = True):
        """
        Reset the hidden states, if the contained module is a recurrent network.

        Args:
            indices: Optionally a sequence of integers or a sequence of
                booleans, specifying which networks' states will be
                reset. If left as None, then the states of all the networks
                will be reset.
            copy: When `indices` is given as something other than None,
                if `copy` is given as True, then the resetting will NOT
                be done in-place. Instead, a new copy of the hidden state
                will first be created, and then the specified regions
                of this new copy will be cleared, and then finally this
                modified copy will be declared as the new hidden state.
                It is a common practice for recurrent neural network
                implementations to return the same tensor both as its
                output and as (part of) its hidden state. With `copy=False`,
                the resetting would be done in-place, and the action
                tensor could be involuntarily reset as well.
                This in-place modification could cause silent bugs
                if the unintended modification on the action tensor
                happens BEFORE the action is sent to the reinforcement
                learning environment.
                To prevent such situations, the default value for the argument
                `copy` is True.
        """
        if indices is None:
            self.__state = None
        else:
            if self.__state is not None:
                with torch.no_grad():
                    if copy:
                        self.__state = deepcopy(self.__state)
                    reset_tensors(self.__state, indices)

    @property
    def parameters(self) -> torch.Tensor:
        """
        The currently used parameters.
        """
        return self.__parameters

    @property
    def h(self) -> Optional[torch.Tensor]:
        """
        The hidden state of the contained recurrent network, if any.

        If the contained recurrent network did not generate a hidden state
        yet, or if the contained network is not recurrent, then the result
        will be None.
        """
        return self.__state

    @property
    def parameter_length(self) -> int:
        """
        Length of the parameter tensor.
        """
        return self.__fmodule.parameter_length

    @property
    def wrapped_module(self) -> nn.Module:
        """
        The wrapped `torch.nn.Module` instance.
        """
        return self.__module

    def to_torch_module(self, parameter_vector: torch.Tensor) -> nn.Module:
        """
        Get a copy of the contained network, parameterized as specified.

        Args:
            parameter_vector: The parameters to be used by the new network.
        Returns:
            Copy of the contained network, as a `torch.nn.Module` instance.
        """
        with torch.no_grad():
            net = deepcopy(self.__module).to(parameter_vector.device)
            nnu.vector_to_parameters(parameter_vector, net.parameters())
        return net



# Copyright 2022 NNAISENSE SA
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import torch

from evotorch import Problem
from evotorch.algorithms import SNES
from evotorch.logging import StdOutLogger


# Define the cost function that receives a batch of solution vectors (2D Tensor) and returns the costs (1D Tensor)
def cost(x: torch.Tensor) -> torch.Tensor:
    return torch.linalg.vector_norm(x, dim=-1)


# The problem is to minimize the cost, for solution vectors of default dtype (float32) and length 10, on the first GPU
problem = Problem(
    "min",
    cost,
    solution_length=10,
    initial_bounds=(-10, 10),
    vectorized=True,
    device="cuda:0" if torch.cuda.is_available() else "cpu",  # Change to "cpu" if a GPU is not available
)

# Run the distribution-based search algorithm SNES for 500 steps and log results to standard output every 100 steps
searcher = SNES(problem, popsize=100, stdev_init=0.01)
logger = StdOutLogger(searcher, interval=100)
searcher.run(500)
print("Final status:\n", searcher.status)



# Copyright 2022 NNAISENSE SA
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Any

import torch
from torch import nn


class StatefulModule(nn.Module):
    """
    A wrapper that provides a stateful interface for recurrent torch modules.

    If the torch module to be wrapped is non-recurrent and its forward method
    has a single input (the input tensor) and a single output (the output
    tensor), then this wrapper module acts as a no-op wrapper.

    If the torch module to be wrapped is recurrent and its forward method has
    two inputs (the input tensor and an optional second argument for the hidden
    state) and two outputs (the output tensor and the new hidden state), then
    this wrapper brings a new forward-passing interface. In this new interface,
    the forward method has a single input (the input tensor) and a single
    output (the output tensor). The hidden states, instead of being
    explicitly requested via a second argument and returned as a second
    result, are stored and used by the wrapper.
    When a new series of inputs is to be used, one has to call the `reset()`
    method of this wrapper.
    """

    def __init__(self, wrapped_module: nn.Module):
        """
        `__init__(...)`: Initialize the StatefulModule.

        Args:
            wrapped_module: The `torch.nn.Module` instance to wrap.
        """
        super().__init__()

        # Declare the variable that will store the hidden state of wrapped_module, if any.
        self._hidden: Any = None

        # Store the module that is wrapped.
        self.wrapped_module = wrapped_module

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        if self._hidden is None:
            # If there is no stored hidden state, then only pass the input tensor to the wrapped module.
            out = self.wrapped_module(x)
        else:
            # If there is a hidden state saved from the previous call to this `forward(...)` method, then pass the
            # input tensor and this stored hidden state.
            out = self.wrapped_module(x, self._hidden)

        if isinstance(out, tuple):
            # If the result of the wrapped module is a tuple, then we assume that the wrapped module returned an
            # output tensor and a hidden state. We assume the first element of this tuple as the output tensor,
            # and the second element as the new hidden state.
            # We set the variable y to the output tensor, and we store the new hidden state via the attribute
            # `_hidden`.
            y, self._hidden = out
        else:
            # If the result of the wrapped module is not a tuple, then we assume that the wrapped module returned
            # only the output tensor. We set the variable y to the output tensor, and set the attribute `_hidden`
            # as None to indicate that there was no hidden state received.
            y = out
            self._hidden = None

        # We return y, which stores the output received by the wrapped module.
        return y

    def reset(self):
        """
        Reset the hidden state, if any.
        """
        self._hidden = None


def ensure_stateful(net: nn.Module):
    """
    Ensure that a module is wrapped by StatefulModule.

    If the given module is already wrapped by StatefulModule, then the
    module itself is returned.
    If the given module is not wrapped by StatefulModule, then this function
    first wraps the module via a new StatefulModule instance, and then this
    new wrapper is returned.

    Args:
        net: The `torch.nn.Module` to be wrapped by StatefulModule (if it is
            not already wrapped by it).
    Returns:
        The module `net`, wrapped by StatefulModule.
    """
    if not isinstance(net, StatefulModule):
        return StatefulModule(net)



# Copyright 2022 NNAISENSE SA
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from evotorch.algorithms import PGPE
from evotorch.logging import PicklingLogger, StdOutLogger
from evotorch.neuroevolution import GymNE

# Specialized Problem class for Gym environments
problem = GymNE(
    env="Humanoid-v4",
    # Linear policy defined using special string syntax supported by EvoTorch
    network="Linear(obs_length, act_length)",
    observation_normalization=True,
    # Humanoid has an "alive bonus" reward at each step, which often hurts EAs, so we remove it
    decrease_rewards_by=5.0,
    # Use all available CPU cores
    num_actors="max",
)

# Run PGPE with ClipUp for 500 steps and log results to standard output at every step
searcher = PGPE(
    problem,
    popsize=200,
    center_learning_rate=0.01125,
    stdev_learning_rate=0.1,
    optimizer="clipup",
    optimizer_config={"max_speed": 0.015},
    radius_init=0.27,
    num_interactions=150000,
    popsize_max=3200,
)

# Instantiate a logger that will print the progress to the standard output
logger = StdOutLogger(searcher)

# Optional:
# Instantiate a logger that will, at every 10 generations, pickle and save the results (where the results will include
# the center of the search distribution since we are using PGPE which is a distribution-based search algorithm).
pickler = PicklingLogger(searcher, interval=10)

# Run the search algorithm
searcher.run(500)

# Create a policy to test using the final center of the optimized distribution and visualize its behavior
population_center = searcher.status["center"]
policy = problem.to_policy(population_center)
problem.visualize(policy)



# Copyright 2022 NNAISENSE SA
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from typing import Iterable

import numpy as np
import torch

from evotorch.neuroevolution.net.runningnorm import RunningNorm
from evotorch.neuroevolution.net.runningstat import RunningStat
from evotorch.testing import assert_allclose


def _compute_stats(observations: Iterable) -> dict:
    # Compute stats from a collection of observations.
    # The observations are expected as a 2-dimensional tensor or numpy array.

    if isinstance(observations, np.ndarray):

        def convert(x):
            return x.numpy()

    else:

        def convert(x):
            return x

    observations = torch.as_tensor(observations, dtype=torch.float32)

    # Get the sum of observations, sum of squares, and the number of observations
    sum_of_obs = torch.sum(observations, dim=0)
    sum_of_squares = torch.sum(observations**2, dim=0)
    count = observations.shape[0]

    # Compute E[x] and E[x^2]
    E_x = sum_of_obs / count
    E_x2 = sum_of_squares / count

    # Compute the variance as E[x^2] - E[x]^2
    # Additionally, clip the variance so that it cannot go below 1e-2.
    variance = E_x2 - (E_x**2)
    variance = torch.max(variance, torch.tensor(1e-2, dtype=torch.float32))

    # The mean is E[x] and the standard deviation is the square root of the variance.
    mean = E_x
    stdev = torch.sqrt(variance)

    return dict(
        sum=convert(sum_of_obs),
        sum_of_squares=convert(sum_of_squares),
        count=count,
        mean=convert(mean),
        stdev=convert(stdev),
    )


def test_running_stat():
    tolerance = 1e-4

    # Create some random observations
    all_obs = np.random.randn(10, 5)

    # Get the stats of the observations
    stats = _compute_stats(all_obs)
    mean = stats["mean"]
    stdev = stats["stdev"]

    # Now that we know the mean and the standard deviation of our observations, we instantiate a RunningStat,
    # feed observations to it, and confirm that it also reports the same mean and the same standard deviation.
    rs = RunningStat()

    for obs in all_obs:
        # Feed each observation to the RunningStat instance.
        rs.update(obs)

    # Confirm that the mean and the standard deviation are reported correctly.
    assert_allclose(rs.mean, mean, atol=tolerance)
    assert_allclose(rs.stdev, stdev, atol=tolerance)

    # We now instantiate multiple RunningStat instances: rs1, rs2, and rs3. Each of them will get its own portion of
    # the observations. Then, we will add each of these RunningStat instances to yet another RunningStat.
    # Finally, we will test whether or not this last RunningStat, after collecting the data from rs1, rs2, and rs3,
    # will report the mean and the standard deviation correctly.
    rs1 = RunningStat()
    rs2 = RunningStat()
    rs3 = RunningStat()

    obs_batch1 = all_obs[:3]
    obs_batch2 = all_obs[3:7]
    obs_batch3 = all_obs[7:]

    for obs in obs_batch1:
        rs1.update(obs)

    for obs in obs_batch2:
        rs2.update(obs)

    for obs in obs_batch3:
        rs3.update(obs)

    final_rs = RunningStat()
    final_rs.update(rs3)
    final_rs.update(rs2)
    final_rs.update(rs1)

    # Assert that `final_rs` reports the correct mean and standard deviation
    assert_allclose(final_rs.mean, mean, atol=tolerance)
    assert_allclose(final_rs.stdev, stdev, atol=tolerance)


def _make_random_booleans(*shape):
    x = torch.randn(*shape)
    return x > 0


def test_running_norm():
    tolerance = 1e-4

    # We imagine an environment whose observation length is as follows:
    obs_length = 10

    # We imagine a batch of environments
    num_envs = 50

    # We will have this many batch of observations
    num_timesteps = 30

    # For each timestep, we generate a batch of observations
    all_obs = torch.randn(num_timesteps, num_envs, obs_length)

    # For each timestep, we generate a mask tensor
    # This mask tensor will tell us which environment is active at that timestep
    masks = _make_random_booleans(num_timesteps, num_envs)

    # Now, we compute the overall stats over all observations over all 'active' environments over all timesteps
    overall_stats = _compute_stats(all_obs.reshape(-1, obs_length)[masks.reshape(-1)])
    overall_mean = overall_stats["mean"]
    overall_stdev = overall_stats["stdev"]

    # Make a new RunningNorm instance
    rn = RunningNorm(shape=obs_length, dtype=torch.float32)

    for t in range(num_timesteps):
        # For each timestep, feed the batch of observations and the mask tensor to the normalizer
        obs = all_obs[t]
        active = masks[t]
        rn.update(obs, mask=active)

    # Assert that the reported mean and stdev of the RunningNorm instance are correct
    assert_allclose(rn.mean, overall_mean, atol=tolerance)
    assert_allclose(rn.stdev, overall_stdev, atol=tolerance)



# Copyright 2022 NNAISENSE SA
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from collections import namedtuple
from copy import deepcopy
from typing import Iterable, Optional, Union

import torch
from torch import nn

from evotorch.tools import Device, DType, to_torch_dtype

CollectedStats = namedtuple("CollectedStats", ["mean", "stdev"])


def _clamp(x: torch.Tensor, min: Optional[float], max: Optional[float]) -> torch.Tensor:
    """
    Clamp the tensor x according to the given min and max values.
    Unlike PyTorch's clamp, this function allows both min and max
    to be None, in which case no clamping will be done.

    Args:
        x: The tensor subject to the clamp operation.
        min: The minimum value.
        max: The maximum value.
    Returns:
        The result of the clamp operation, as a tensor.
        If both min and max were None, the returned object is x itself.
    """
    if (min is None) and (max is None):
        return x
    else:
        return torch.clamp(x, min, max)


class RunningNorm:
    """
    An online observation normalization tool
    """

    def __init__(
        self,
        *,
        shape: Union[tuple, int],
        dtype: DType,
        device: Optional[Device] = None,
        min_variance: float = 1e-2,
        clip: Optional[tuple] = None,
    ) -> None:
        """
        `__init__(...)`: Initialize the RunningNorm

        Args:
            shape: Observation shape. Can be an integer or a tuple.
            dtype: The dtype of the observations.
            device: The device in which the observation stats are held.
                If left as None, the device is assumed to be "cpu".
            min_variance: A lower bound for the variance to be used in
                the normalization computations.
                In other words, if the computed variance according to the
                collected observations ends up lower than `min_variance`,
                this `min_variance` will be used instead (in an elementwise
                manner) while computing the normalized observations.
                As in Salimans et al. (2017), the default is 1e-2.
            clip: Can be left as None (which is the default), or can be
                given as a pair of real numbers.
                This is used for clipping the observations after the
                normalization operation.
                In Salimans et al. (2017), (-5.0, +5.0) was used.
        """

        # Make sure that the shape is stored as a torch.Size object.
        if isinstance(shape, Iterable):
            self._shape = torch.Size(shape)
        else:
            self._shape = torch.Size([int(shape)])

        # Store the number of dimensions
        self._ndim = len(self._shape)

        # Store the dtype and the device
        self._dtype = to_torch_dtype(dtype)
        self._device = "cpu" if device is None else device

        # Initialize the internally stored data as empty
        self._sum: Optional[torch.Tensor] = None
        self._sum_of_squares: Optional[torch.Tensor] = None
        self._count: int = 0

        # Store the minimum variance
        self._min_variance = float(min_variance)

        if clip is not None:
            # If a clip tuple was provided, store the specified lower and upper bounds
            lb, ub = clip
            self._lb = float(lb)
            self._ub = float(ub)
        else:
            # If a clip tuple was not provided the bounds are stored as None
            self._lb = None
            self._ub = None

    def to(self, device: Device) -> "RunningNorm":
        """
        If the target device is a different device, then make a copy of this
        RunningNorm instance on the target device.
        If the target device is the same with this RunningNorm's device, then
        return this RunningNorm itself.

        Args:
            device: The target device.
        Returns:
            The RunningNorm on the target device. This can be a copy, or the
            original RunningNorm instance itself.
        """
        if torch.device(device) == torch.device(self.device):
            return self
        else:
            new_running_norm = object.__new__(type(self))

            already_handled = {"_sum", "_sum_of_squares", "_device"}
            new_running_norm._sum = self._sum.to(device)
            new_running_norm._sum_of_squares = self._sum_of_squares.to(device)
            new_running_norm._device = device

            for k, v in self.__dict__.items():
                if k not in already_handled:
                    setattr(new_running_norm, k, deepcopy(v))

            return new_running_norm

    @property
    def device(self) -> Device:
        """
        The device in which the observation stats are held
        """
        return self._device

    @property
    def dtype(self) -> DType:
        """
        The dtype of the stored observation stats
        """
        return self._dtype

    @property
    def shape(self) -> tuple:
        """
        Observation shape
        """
        return self._shape

    @property
    def min_variance(self) -> float:
        """
        Minimum variance
        """
        return self._min_variance

    @property
    def low(self) -> Optional[float]:
        """
        The lower component of the bounds given in the `clip` tuple.
        If `clip` was initialized as None, this is also None.
        """
        return self._lb

    @property
    def high(self) -> Optional[float]:
        """
        The higher (upper) component of the bounds given in the `clip` tuple.
        If `clip` was initialized as None, this is also None.
        """
        return self._ub

    def _like_its_own(self, x: Iterable) -> torch.Tensor:
        return torch.as_tensor(x, dtype=self._dtype, device=self._device)

    def _verify(self, x: Iterable) -> torch.Tensor:
        x = self._like_its_own(x)
        if x.ndim == self._ndim:
            if x.shape != self._shape:
                raise ValueError(
                    f"This RunningNorm instance was initialized with shape: {self._shape}."
                    f" However, the provided tensor has an incompatible shape: {x._shape}."
                )
        elif x.ndim == (self._ndim + 1):
            if x.shape[1:] != self._shape:
                raise ValueError(
                    f"This RunningNorm instance was initialized with shape: {self._shape}."
                    f" The provided tensor is shaped {x.shape}."
                    f" Accepting the tensor's leftmost dimension as the batch size,"
                    f" the remaining shape is incompatible: {x.shape[1:]}"
                )
        else:
            raise ValueError(
                f"This RunningNorm instance was initialized with shape: {self._shape}."
                f" The provided tensor is shaped {x.shape}."
                f" The number of dimensions of the given tensor is incompatible."
            )
        return x

    def _has_no_data(self) -> bool:
        return (self._sum is None) and (self._sum_of_squares is None) and (self._count == 0)

    def _has_data(self) -> bool:
        return (self._sum is not None) and (self._sum_of_squares is not None) and (self._count > 0)

    def reset(self):
        """
        Remove all the collected observation data.
        """
        self._sum = None
        self._sum_of_squares = None
        self._count = 0

    @torch.no_grad()
    def update(self, x: Union[Iterable, "RunningNorm"], mask: Optional[Iterable] = None, *, verify: bool = True):
        """
        Update the stored stats with new observation data.

        Args:
            x: The new observation(s), as a PyTorch tensor, or any Iterable
                that can be converted to a PyTorch tensor, or another
                RunningNorm instance.
                If given as a tensor or as an Iterable, the shape of `x` can
                be the same with observation shape, or it can be augmented
                with an extra leftmost dimension.
                In the case of augmented dimension, `x` is interpreted not as
                a single observation, but as a batch of observations.
                If `x` is another RunningNorm instance, the stats stored by
                this RunningNorm instance will be updated with all the data
                stored by `x`.
            mask: Can be given as a 1-dimensional Iterable of booleans ONLY
                if `x` represents a batch of observations.
                If a `mask` is provided, the i-th observation within the
                observation batch `x` will be taken into account only if
                the i-th item of the `mask` is True.
            verify: Whether or not to verify the shape of the given Iterable
                objects. The default is True.
        """
        if isinstance(x, RunningNorm):
            # If we are to update our stats according to another RunningNorm instance

            if x._count > 0:
                # We bother only if x is non-empty

                if mask is not None:
                    # We were given another RunningNorm, not a batch of observations.
                    # So, we do not expect to receive a mask tensor.
                    # If a mask was provided, then this is an unexpected way of calling this function.
                    # We therefore raise an error.
                    raise ValueError(
                        "The `mask` argument is expected as None if the first argument is a RunningNorm."
                        " However, `mask` is found as something other than None."
                    )

                if self._shape != x._shape:
                    # If the shapes of this RunningNorm and of the other RunningNorm
                    # do not match, then we cannot use `x` for updating our stats.
                    # It might be the case that `x` was initialized for another
                    # task, with differently sized observations.
                    # We therefore raise an error.
                    raise ValueError(
                        f"The RunningNorm to be updated has the shape {self._shape}"
                        f" The other RunningNorm has the shape {self._shape}"
                        f" These shapes are incompatible."
                    )

                if self._has_no_data():
                    # If this RunningNorm has no data at all, then we clone the
                    # data of x.
                    self._sum = self._like_its_own(x._sum.clone())
                    self._sum_of_squares = self._like_its_own(x._sum_of_squares.clone())
                    self._count = x._count
                elif self._has_data():
                    # If this RunningNorm has its own data, then we update the
                    # stored data with the data stored by x.
                    self._sum += self._like_its_own(x._sum)
                    self._sum_of_squares += self._like_its_own(x._sum_of_squares)
                    self._count += x._count
                else:
                    assert False, "RunningNorm is in an invalid state! This might be a bug."
        else:
            # This is the case where the received argument x is not a
            # RunningNorm object, but an Iterable.

            if verify:
                # If we have the `verify` flag, then we make sure that
                # x is a tensor of the correct shape
                x = self._verify(x)

            if x.ndim == self._ndim:
                # If the shape of x is exactly the same with the observation shape
                # then we assume that x represents a single observation, and not a
                # batch of observations.

                if mask is not None:
                    # Since we are dealing with a single observation,
                    # we do not expect to receive a mask argument.
                    # If the mask argument was provided, then this is an unexpected
                    # usage of this function.
                    # We therefore raise an error.
                    raise ValueError(
                        "The `mask` argument is expected as None if the first argument is a single observation"
                        " (i.e. not a batch of observations, with an extra leftmost dimension)."
                        " However, `mask` is found as something other than None."
                    )

                # Since x is a single observation,
                # the sum of observations extracted from x is x itself,
                # and the sum of squared observations extracted from x is
                # the square of x itself.
                sum_of_x = x
                sum_of_x_squared = x.square()
                # We extracted a single observation from x
                n = 1
            elif x.ndim == (self._ndim + 1):
                # If the number of dimensions of x is one more than the number
                # of dimensions of this RunningNorm, then we assume that x is a batch
                # of observations.

                if mask is not None:
                    # If a mask is provided, then we first make sure that it is a tensor
                    # of dtype bool in the correct device.
                    mask = torch.as_tensor(mask, dtype=torch.bool, device=self._device)

                    if mask.ndim != 1:
                        # We expect the mask to be 1-dimensional.
                        # If not, we raise an error.
                        raise ValueError(
                            f"The `mask` tensor was expected as a 1-dimensional tensor."
                            f" However, its shape is {mask.shape}."
                        )

                    if len(mask) != x.shape[0]:
                        # If the length of the mask is not the batch size of x,
                        # then there is a mismatch.
                        # We therefore raise an error.
                        raise ValueError(
                            f"The shape of the given tensor is {x.shape}."
                            f" Therefore, the batch size of observations is {x.shape[0]}."
                            f" However, the given `mask` tensor does not has an incompatible length: {len(mask)}."
                        )

                    # We compute how many True items we have in the mask.
                    # This integer gives us how many observations we extract from x.
                    n = int(torch.sum(torch.as_tensor(mask, dtype=torch.int64, device=self._device)))

                    # We now re-cast the mask as the observation dtype (so that True items turn to 1.0
                    # and False items turn to 0.0), and then increase its number of dimensions so that
                    # it can operate directly with x.
                    mask = self._like_its_own(mask).reshape(torch.Size([x.shape[0]] + ([1] * (x.ndim - 1))))

                    # Finally, we multiply x with the mask. This means that the observations with corresponding
                    # mask values as False are zeroed out.
                    x = x * mask
                else:
                    # This is the case where we did not receive a mask.
                    # We can simply say that the number of observations to extract from x
                    # is the size of its leftmost dimension, i.e. the batch size.
                    n = x.shape[0]

                # With or without a mask, we are now ready to extract the sum and sum of squares
                # from x.
                sum_of_x = torch.sum(x, dim=0)
                sum_of_x_squared = torch.sum(x.square(), dim=0)
            else:
                # This is the case where the number of dimensions of x is unrecognized.
                # This case is actually already checked by the _verify(...) method earlier.
                # This defensive fallback case is only for when verify=False and it turned out
                # that the ndim is invalid.
                raise ValueError(f"Invalid shape: {x.shape}")

            # At this point, we handled all the valid cases regarding the Iterable x,
            # and we have our sum_of_x (sum of all observations), sum_of_squares
            # (sum of all squared observations), and n (number of observations extracted
            # from x).

            if self._has_no_data():
                # If our RunningNorm is empty, the observation data we extracted from x
                # become our RunningNorm's new data.
                self._sum = sum_of_x
                self._sum_of_squares = sum_of_x_squared
                self._count = n
            elif self._has_data():
                # If our RunningNorm is not empty, the stored data is updated with the
                # data extracted from x.
                self._sum += sum_of_x
                self._sum_of_squares += sum_of_x_squared
                self._count += n
            else:
                # This is an erroneous state where the internal data looks neither
                # existent nor completely empty.
                # This might be the result of a bug, or maybe this instance's
                # protected variables were tempered with from the outside.
                assert False, "RunningNorm is in an invalid state! This might be a bug."

    @property
    @torch.no_grad()
    def stats(self) -> CollectedStats:
        """
        The collected data's mean and standard deviation (stdev) in a tuple
        """

        # Using the internally stored sum, sum_of_squares, and count,
        # compute E[x] and E[x^2]
        E_x = self._sum / self._count
        E_x2 = self._sum_of_squares / self._count

        # The mean is E[x]
        mean = E_x

        # The variance is E[x^2] - (E[x])^2, elementwise clipped such that
        # it cannot go below min_variance
        variance = _clamp(E_x2 - E_x.square(), self._min_variance, None)

        # Standard deviation is finally computed as the square root of the variance
        stdev = torch.sqrt(variance)

        # Return the stats in a named tuple
        return CollectedStats(mean=mean, stdev=stdev)

    @property
    def mean(self) -> torch.Tensor:
        """
        The collected data's mean
        """
        return self._sum / self._count

    @property
    def stdev(self) -> torch.Tensor:
        """
        The collected data's standard deviation
        """
        return self.stats.stdev

    @property
    def sum(self) -> torch.Tensor:
        """
        The collected data's sum
        """
        return self._sum

    @property
    def sum_of_squares(self) -> torch.Tensor:
        """
        Sum of squares of the collected data
        """
        return self._sum_of_squares

    @property
    def count(self) -> int:
        """
        Number of observations encountered
        """
        return self._count

    @torch.no_grad()
    def normalize(self, x: Iterable, *, result_as_numpy: Optional[bool] = None, verify: bool = True) -> Iterable:
        """
        Normalize the given observation x.

        Args:
            x: The observation(s), as a PyTorch tensor, or any Iterable
                that is convertable to a PyTorch tensor.
                `x` can be a single observation, or it can be a batch
                of observations (with an extra leftmost dimension).
            result_as_numpy: Whether or not to return the normalized
                observation as a numpy array.
                If left as None (which is the default), then the returned
                type depends on x: a PyTorch tensor is returned if x is a
                PyTorch tensor, and a numpy array is returned otherwise.
                If True, the result is always a numpy array.
                If False, the result is always a PyTorch tensor.
            verify: Whether or not to check the type and dimensions of x.
                This is True by default.
                Note that, if `verify` is False, this function will not
                properly check the type of `x` and will assume that `x`
                is a PyTorch tensor.
        Returns:
            The normalized observation, as a PyTorch tensor or a numpy array.
        """

        if self._count == 0:
            # If this RunningNorm instance has no data yet,
            # then we do not know how to do the normalization.
            # We therefore raise an error.
            raise ValueError("Cannot do normalization because no data is collected yet.")

        if verify:
            # Here we verify the type and shape of x.

            if result_as_numpy is None:
                # If there is not an explicit request about the return type,
                # we infer the return type from the type of x:
                # if x is a tensor, we return a tensor;
                # otherwise, we assume x to be a CPU-bound iterable, and
                # therefore we return a numpy array.
                result_as_numpy = not isinstance(x, torch.Tensor)
            else:
                result_as_numpy = bool(result_as_numpy)

            # We call _verify() to make sure that x is of correct shape
            # and is properly converted to a PyTorch tensor.
            x = self._verify(x)

        # We get the mean and stdev of the collected data
        mean, stdev = self.stats

        # Now we compute the normalized observation, clipped according to the
        # lower and upper bounds expressed by the `clip` tuple, if exists.
        result = _clamp((x - mean) / stdev, self._lb, self._ub)

        if result_as_numpy:
            # If we are to return the result as a numpy array, we do the
            # necessary conversion.
            result = result.cpu().numpy()

        # Finally, return the result
        return result

    @torch.no_grad()
    def update_and_normalize(self, x: Iterable, mask: Optional[Iterable] = None) -> Iterable:
        """
        Update the observation stats according to x, then normalize x.

        Args:
            x: The observation(s), as a PyTorch tensor, or as an Iterable
                which can be converted to a PyTorch tensor.
                The shape of x can be the same with the observaiton shape,
                or it can be augmented with an extra leftmost dimension
                to express a batch of observations.
            mask: Can be given as a 1-dimensional Iterable of booleans ONLY
                if `x` represents a batch of observations.
                If a `mask` is provided, the i-th observation within the
                observation batch `x` will be taken into account only if
                the the i-th item of the `mask` is True.
        Returns:
            The normalized counterpart of the observation(s) expressed by x.
        """
        result_as_numpy = not isinstance(x, torch.Tensor)
        x = self._verify(x)

        self.update(x, mask, verify=False)
        result = self.normalize(x, verify=False)

        if result_as_numpy:
            result = result.cpu().numpy()

        return result

    def to_layer(self) -> "ObsNormLayer":
        """
        Make a PyTorch module which normalizes the its inputs.

        Returns:
            An ObsNormLayer instance.
        """
        mean, stdev = self.stats
        low = self.low
        high = self.high
        return ObsNormLayer(mean=mean, stdev=stdev, low=low, high=high)

    def __repr__(self) -> str:
        return f"<{self.__class__.__name__}, count: {self.count}>"

    def __copy__(self) -> "RunningNorm":
        return deepcopy(self)


class ObsNormLayer(nn.Module):
    """
    An observation normalizer which behaves as a PyTorch Module.
    """

    def __init__(
        self, mean: torch.Tensor, stdev: torch.Tensor, low: Optional[float] = None, high: Optional[float] = None
    ) -> None:
        """
        `__init__(...)`: Initialize the ObsNormLayer.

        Args:
            mean: The mean according to which the observations are to be
                normalized.
            stdev: The standard deviation according to which the observations
                are to be normalized.
            low: Optionally a real number if the result of the normalization
                is to be clipped. Represents the lower bound for the clipping
                operation.
            high: Optionally a real number if the result of the normalization
                is to be clipped. Represents the upper bound for the clipping
                operation.
        """
        super().__init__()
        self.register_buffer("_mean", mean)
        self.register_buffer("_stdev", stdev)
        self._lb = None if low is None else float(low)
        self._ub = None if high is None else float(high)

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        """
        Normalize an observation or a batch of observations.

        Args:
            x: The observation(s).
        Returns:
            The normalized counterpart of the observation(s).
        """
        return _clamp((x - self._mean) / self._stdev, self._lb, self._ub)
